
// @module : dummy_core
// @author : Michel Kinsy
 /* 
 *  Using Heracles Multicore System under the copytright below:
 *  Copyright (c) 2011 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 *
 */
 
//----------------------------------------------------
// Parameterized dummy core.
//----------------------------------------------------

module dummy_core #(parameter TRAFFIC = 1, RATE = 5, 
							  DATA_WIDTH = 32, IN_PORTS  = 1, OUT_PORTS  = 1, VC_PER_PORTS = 2,
							  ID_BITS = 4, EXTRA  = 2, TYPE_BITS = 2, ROW = 3, COLUMN = 3, 
							  FLITS_PER_PACKET = 8, STATS = 1024) (
		clock, reset, start, stop, 
		core_ID, dest_ID, 

		from_peripheral,
		from_peripheral_data, 
		from_peripheral_valid, 
		to_peripheral,
		to_peripheral_data, 
		to_peripheral_valid,	 

		from_core_flit, v_from_core, 
		from_core_empty, from_core_full,

		to_core_flit, v_to_core,
		to_core_empty, to_core_full,
		// performance reporting
		report
);

  //define the log2 function
	function integer log2;
	  input integer num;
	  integer i, result;
	  begin
		for (i = 0; 2 ** i < num; i = i + 1)
		  result = i + 1;
		log2 = result;
	  end
	endfunction

  //define shuffle function
	function integer suffle;
	  input integer num;
	  integer i, temp, result;
	  begin
	    temp = ((ROW * COLUMN) -1) / 2; 
		if (num > temp)
			result = (((num - temp) -1) <<1) + 1; 
		else 
			result = num << 1; 
		suffle = result;
	  end
	endfunction

localparam VC_BITS = log2(VC_PER_PORTS); 
localparam FLOW_BITS = (2*ID_BITS) + EXTRA;  
//Number of bits to represent coordinates. 
localparam ROW_BITS = log2(ROW); 
localparam COLUMN_BITS = log2(COLUMN);
  
localparam NODE_BITS = ROW_BITS + COLUMN_BITS; // should be the same as ID_BITS 
localparam IN_VCS    = (VC_PER_PORTS * IN_PORTS); 
localparam FLIT_WIDTH    = FLOW_BITS + TYPE_BITS + VC_BITS + DATA_WIDTH;

localparam DEF_T = 0, BIT_COMP = 1, TRANS = 2, SHUFFLE = 3; 

input clock, reset, start, stop; 
input [ID_BITS - 1: 0] core_ID; 
input [ID_BITS - 1: 0] dest_ID;

input  [1:0]   from_peripheral;
input  [31:0]  from_peripheral_data; 
input          from_peripheral_valid;
output [1:0]   to_peripheral;
output [31:0]  to_peripheral_data; 
output         to_peripheral_valid;

output [(FLIT_WIDTH * OUT_PORTS) -1:0] from_core_flit;
output [OUT_PORTS-1: 0] v_from_core;
output [IN_VCS -1:0] from_core_empty;
output [IN_VCS -1:0] from_core_full; 
  
input  [(FLIT_WIDTH * OUT_PORTS) -1:0] to_core_flit;
input  [OUT_PORTS-1: 0] v_to_core;
input  [IN_VCS -1:0] to_core_empty;
input  [IN_VCS -1:0] to_core_full;

output report; 

reg  [1:0]     to_peripheral;
reg  [31:0]    to_peripheral_data; 
reg            to_peripheral_valid;

  function [ID_BITS:0] grid_id;
	input integer id; 
	integer i, j, count; 
	reg [ROW_BITS - 1:0] row; reg [COLUMN_BITS - 1:0] col; 
	begin
        count = 0; row = 0; col = 0; 
		for(i= 0; i < ROW; i = i+1) begin
			for(j= 0; j < COLUMN; j = j+1) begin
				if (count == id) begin
					row = i; 
					col = j; 
				end
				count = count +1; 
			end
		end  
		grid_id = {row, col}; 
	end
  endfunction
  
  function integer get_id;
	input [ID_BITS:0] id; 
	integer i, j, node, count; 
	reg [ROW_BITS - 1:0] row; reg [COLUMN_BITS - 1:0] col; 
	begin
        count = 0; 
		node  = 0; 
		row = id[(NODE_BITS - 1)-: ROW_BITS]; 
		col = id[(COLUMN_BITS - 1): 0]; 
		for(i= 0; i < ROW; i = i+1) begin
			for(j= 0; j < COLUMN; j = j+1) begin
				if ((i == row) & (j == col)) begin
					node = count; 
				end
				count = count+1; 
			end
		end  
		get_id = node; 
	end
  endfunction
 
wire [ID_BITS - 1: 0] core_coord = grid_id(core_ID); 
wire [ID_BITS - 1: 0] dest_coord = grid_id(dest_ID); 

wire [(COLUMN_BITS - 1): 0] cur_x = core_coord[(COLUMN_BITS - 1): 0]; 
wire [(ROW_BITS - 1): 0]    cur_y = core_coord[(NODE_BITS - 1)-: ROW_BITS];

wire [ID_BITS - 1: 0] dest = (TRAFFIC == BIT_COMP)? ~core_coord : 
							    (TRAFFIC == TRANS)?    {cur_x, cur_y} : 
							    (TRAFFIC == SHUFFLE)?  suffle(core_coord) : 
							    dest_coord;
  
wire [(COLUMN_BITS - 1): 0] dest_x = dest[(COLUMN_BITS - 1): 0]; 
wire [(ROW_BITS - 1): 0]    dest_y = dest[(NODE_BITS - 1)-: ROW_BITS];

reg [IN_VCS -1:0] from_core_full  = 0;  // infinite capacity sink
reg [IN_VCS -1:0] from_core_empty = ~0;

reg [FLIT_WIDTH -1:0] t_from_core_flit [0 : OUT_PORTS-1];
reg [IN_PORTS-1: 0]   t_v_from_core;
  
wire  [FLIT_WIDTH -1:0] t_to_core_flit [0 : IN_PORTS-1];
wire  [VC_PER_PORTS -1:0] t_to_core_empty [0 : IN_PORTS-1];
wire  [VC_PER_PORTS -1:0] t_to_core_full [0 : IN_PORTS-1];

wire [TYPE_BITS - 1:0] r_type [0 : IN_PORTS-1];
wire [ID_BITS - 1: 0] source = core_coord; 
wire [ID_BITS - 1: 0] destination = (~stop)? dest : destination; 
reg  [EXTRA - 1: 0]   sub_flow = 0; 

wire [ID_BITS - 1:0] r_sender [0 : IN_PORTS-1];
reg [ID_BITS - 1:0]  last_sender; 

reg [FLIT_WIDTH - 1:0] flit;
reg [FLOW_BITS - 1:0]  flow_ID; 
reg [TYPE_BITS - 1:0]  flit_type;
reg [VC_BITS - 1:0]    flit_vc;
reg [DATA_WIDTH - 1:0] flit_payload; 

reg [log2(RATE)    : 0] injection_rate;
reg [31 : 0]            packet_injected; 
reg [31 : 0]            packet_recieved;

reg [log2(FLITS_PER_PACKET) : 0]  counter;

reg [DATA_WIDTH - 1:0]  data; 
localparam HEAD = 'b10, BODY = 'b00, TAIL = 'b01, ALL= 'b11;
localparam IDLE = 0, SEARCH_ST = 1, DISPATCH = 2, FILL_ST = 3, SENT = 4; 
localparam NONE = 0, P_REC = 1, P_INJ = 2;
reg [2: 0] state; 
reg [1: 0] reporting; 
  
reg  found; 
reg [VC_BITS - 1:0]    free_vc; 
reg [IN_PORTS-1 :0]    free_port; 
reg [63: 0] run_cycles;
  
integer index, j;
assign report = ((stop & (reporting == NONE)) | (run_cycles == STATS)); 

genvar i;
 generate
      	for (i = 0; i < IN_PORTS; i = i+1) begin : IN_CPORTS
      	 	assign t_to_core_flit[i]    = to_core_flit[((FLIT_WIDTH *(i + 1))-1) -: FLIT_WIDTH];
      	 	assign t_to_core_empty[i]   = to_core_empty[((VC_PER_PORTS *(i + 1))-1) -: VC_PER_PORTS]; 
      	 	assign t_to_core_full[i]    = to_core_full[((VC_PER_PORTS *(i + 1))-1) -: VC_PER_PORTS]; 
      	end  
      	
      	for (i = 0; i < OUT_PORTS; i = i+1) begin : OUT_CPORTS
			   assign from_core_flit [(((i + 1) *(FLIT_WIDTH))-1) -: FLIT_WIDTH]  = t_from_core_flit[i];
      	end 	
      	assign v_from_core = t_v_from_core;
		
      	for (i = 0; i < IN_PORTS; i = i+1) begin : IN_TYPE
			assign r_type[i]   = t_to_core_flit[i] >> (VC_BITS + DATA_WIDTH);
			assign r_sender[i] = t_to_core_flit[i][(FLIT_WIDTH-1) -: ID_BITS]; 
      	end  
 endgenerate 

 always @ (posedge clock) begin          // ----- Injection logic 
   if(reset == 1) begin 
	  injection_rate  <= 0; 
	  packet_injected <= 0; 
	  data            <= 1;
      state           <= IDLE; 
      for(index= 0; index < IN_PORTS; index = index+1) begin
          t_from_core_flit[index]  <= 0;
	      t_v_from_core[index]     <= 0; 
      end
  end 
  else begin
	if (start) begin
		case (state)
			IDLE: begin
				 t_from_core_flit[free_port]  <= 0;
				 t_v_from_core[free_port]     <= 0;
				 
				 counter   <= 0;
				 found     <= 0; 
				 free_vc   <= 0; 
				 free_port <= 0; 
				 injection_rate <= injection_rate + 1; 
				 if(injection_rate == RATE) begin 
					state     <= SEARCH_ST;
				 end 
				 else begin 
					state     <= IDLE;
				end 
			end
			SEARCH_ST: begin
				for(index= 0; index < IN_PORTS; index = index+1) begin
					for(j= 0; j < VC_PER_PORTS; j = j+1) begin 
						if(t_to_core_empty [index][j] & ~found) begin
						   free_vc     <= j;
						   free_port   <= index;
						   found       <= 1; 
						end 
					end
				end
				if(found)  state     <= DISPATCH;
				else state           <= SEARCH_ST;
			end

			DISPATCH: begin
				packet_injected<= packet_injected + 1; 
				flit_type      <= (FLITS_PER_PACKET == 1)? ALL : HEAD;
				flit_vc        <= free_vc;
				flit_payload   <= data;  
				data           <= data + 1; 
				counter        <= counter + 1;
				state          <= SENT;         	     		       
			end

			FILL_ST: begin
				t_v_from_core[free_port]     <= 0;
				if (~(t_to_core_full[free_port][free_vc])) begin 
					flit_type      <= (FLITS_PER_PACKET == 1)? ALL : 
									  (counter == (FLITS_PER_PACKET-1))? TAIL : BODY;
					flit_payload   <= data;   
					data           <= data + 1; 
					counter        <= counter + 1;
					state          <= SENT; 
				end 
				else begin 
					state          <= FILL_ST;
				end 
			end  

			SENT: begin
				if (~(t_to_core_full[free_port][free_vc])) begin 
					t_from_core_flit[free_port]  <= {source, destination, sub_flow, flit_type, flit_vc, flit_payload};
					t_v_from_core[free_port]     <= 1;  
					state                        <= ((flit_type == TAIL) | (flit_type == ALL))? IDLE : FILL_ST;    
					//$display (" Core %d sending [%h] on vc [%b]", core_ID, {source, destination, 
					//sub_flow, flit_type, flit_vc, flit_payload}, flit_vc);
				end 
				else begin 
					t_v_from_core[free_port]     <= 0; 
					state                        <= SENT;
				end 
			end 
			default:    state   <= IDLE;
		endcase 
	end 
	else begin 
		  state        <= IDLE; 
		  for(index= 0; index < IN_PORTS; index = index+1) begin
			  t_from_core_flit[index]  <= 0;
			  t_v_from_core[index]     <= 0; 
		  end
	end 
   end 
end 

always @ (posedge clock) begin           // ----- Recieving logic 
      if (reset) begin 
			packet_recieved = 0;
			last_sender     = 0;
      end
      else begin  
			if (~stop) begin	  
				for(index= 0; index < IN_PORTS; index = index+1) begin 
					 if(v_to_core[index]) begin
						//$display (" Core %d recieved [%h] on vc [%b]", core_ID, t_to_core_flit[index], 
						//t_to_core_flit[index][(FLIT_WIDTH - FLOW_BITS - TYPE_BITS -1) -: VC_BITS]);						
						if((r_type[index] == HEAD) | (r_type[index] == ALL))begin 
							packet_recieved =  packet_recieved + 1;
							last_sender     =  r_sender[index]; 
						end
					 end
				end 
			end
      end
end 

always @ (posedge clock) begin           // ----- Reporting logic 
      if (reset) begin 
		 run_cycles  <= 0; 
		 reporting      <= NONE; 
      end
      else begin 
            if (stop) begin 
				if (reporting == NONE) begin 
					to_peripheral       <= 0;
					to_peripheral_data  <= packet_recieved ; 
					to_peripheral_valid <= 1;
					reporting              <= P_REC; 
				end 
				else begin 
					if (reporting == P_REC) begin 
						to_peripheral       <= 0;
						to_peripheral_data  <= packet_injected ; 
						to_peripheral_valid <= 1;
						reporting              <= P_INJ; 
					end 
					else begin 
						to_peripheral_valid <= 0; 
					end 
				end 	
			end 
			else begin 
				run_cycles  <= run_cycles + 1;
				to_peripheral_valid <= 0; 				
			end 
			
			if((stop & (reporting == NONE)) | (run_cycles == STATS))begin 
				$display ("---------------------------------- Dummy Core %d ----------------------------------", core_ID);  
				$display ("Total run cycles       [%d]", run_cycles);
				$display ("Injection rate         [%d]", RATE);
				$display ("Destination            [%d]", destination);
				$display ("Total injected packets [%d]", packet_injected);
				$display ("Last sender            [%d]", get_id(last_sender));
				$display ("Total recieved packets [%d]", packet_recieved);
				$display ("----------------------------------------------------------------------------------");
			end 
      end
end

/*
 always @ (posedge clock) begin 
		$display ("---------------------------------- Dummy Core %d ----------------------------------", core_ID);  
		$display ("Reset [%b]\t\t| State [%d]", reset, state);

		$display ("free_vc[%b]\t\t| free_port  [%b]\t| found [%b]",free_vc, free_port , found);
		$display ("counter[%d]\t\t| destination [%b]",counter, destination);
		$display ("flit_type[%d]\t\t| flit_vc [%b]\t\t| flit_payload [%h]",flit_type, flit_vc , flit_payload);
		
		$display ("From Core Flit [%h] ", from_core_flit);
		$display ("From Core Valid [%b] ", v_from_core);
		$display ("From Core Empty [%b]\t| From Core Full [%b] ", from_core_empty, from_core_full);
		 
		$display ("To Core Flit [%h] ", to_core_flit);
		$display ("To Core Valid [%b] ", v_to_core);
		$display ("To Core Empty [%b]\t| To Core Full [%b] ", to_core_empty, to_core_full);
		$display ("Inject_Rate [%d] Inject_Count [%d] Reciev_Count [%d]", injection_rate,packet_injected,packet_recieved);
		$display ("----------------------------------------------------------------------------------"); 
 end 	
//*/ 

endmodule 
