
/** @module : Packetizer_Lite
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2010 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
//localparam EXTRA  = 2; // this allows multiple flows to be identified between one pair of nodes. 

module packetizer_lite #(parameter CORE = 0, DATA_WIDTH = 32, ADDRESS_BITS = 20, OFFSET_BITS= 3, 
                         IN_PORTS  = 1, OUT_PORTS  = 1, VC_PER_PORTS = 2, VC_BITS = 1,
					     VC_DEPTH_BITS = 4, ID_BITS = 4, EXTRA  = 2, TYPE_BITS = 2) (
				     clock, reset,       				     
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full,  
				     
				     in_address, 
         		  	 in_cache_line,
         		     in_valid,  
            		
         		  	 out_address,
         		  	 out_cache_line,
         		  	 out_valid,
         		  	 reply2cache, 
            		 
         		  	 network_req_address, 
         		 	 network_req,
            		 
				     ready, read, write, ack, 
				     address, mem_in_cache_line,
				     mem_valid, mem_out_cache_line		 
);

localparam FLOW_BITS      = (2*ID_BITS) + EXTRA;  
localparam FLIT_WIDTH     = FLOW_BITS + TYPE_BITS + VC_BITS + DATA_WIDTH; 
localparam REAL_ADDR_BITS = ADDRESS_BITS - ID_BITS;
localparam OFFSET         = (1<<OFFSET_BITS); 

input clock, reset; 

output [(FLIT_WIDTH * OUT_PORTS) -1:0] from_core_flit;
output [OUT_PORTS-1: 0] v_from_core;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_empty;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_full; 
  
input  [(FLIT_WIDTH * OUT_PORTS) -1:0] to_core_flit;
input  [OUT_PORTS-1: 0] v_to_core;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_empty;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_full;

output [ADDRESS_BITS-1:0] in_address;
output [(DATA_WIDTH * OFFSET)-1:0]  in_cache_line;
output in_valid;

input [ADDRESS_BITS-1:0] out_address;
input [(DATA_WIDTH * OFFSET)-1:0]  out_cache_line;
input out_valid;
output reply2cache; 
input [ADDRESS_BITS-1:0] network_req_address;
input network_req;

input  ready; 
output read, write, ack;
output [ADDRESS_BITS-1:0] address;
output [(DATA_WIDTH * OFFSET)-1:0]  mem_in_cache_line;
input mem_valid; 
input [(DATA_WIDTH * OFFSET)-1:0]  mem_out_cache_line;

//define the log2 function
function integer log2;
	input integer num;
	integer i, result;
	begin
		result = 1; 
		for (i = 0; 2 ** i < num; i = i + 1)
			result = i + 1;
		log2 = result;
	end
endfunction

localparam HEAD = 'b10, BODY = 'b00, TAIL = 'b01, ALL= 'b11;
localparam IDLE = 0, SEARCH_ST = 1, DISPATCH = 2, READY_ST = 3, REQ_ST = 4, VAL_ST = 5, FILL_ST = 6, SENT = 7; 
localparam R_IDLE = 0, CHECK_VC = 1, READ_DELAY = 2, GET_FLIT = 3, PROCESS = 4, R_LAST = 5;  
localparam PORT_BITS = log2(IN_PORTS);

reg [FLIT_WIDTH -1:0] t_from_core_flit [0 : IN_PORTS-1];
reg [IN_PORTS-1: 0] t_v_from_core;
wire [VC_PER_PORTS -1:0] t_from_core_empty [0 : IN_PORTS-1];
wire [VC_PER_PORTS -1:0] t_from_core_full [0 : IN_PORTS-1]; 
  
wire  [FLIT_WIDTH -1:0] t_to_core_flit [0 : OUT_PORTS-1];
wire  [OUT_PORTS-1: 0] t_v_to_core;
wire  [VC_PER_PORTS -1:0] t_to_core_empty [0 : OUT_PORTS-1];
wire  [VC_PER_PORTS -1:0] t_to_core_full [0 : OUT_PORTS-1];

reg [FLIT_WIDTH - 1:0] in_flit [0:IN_PORTS-1];
reg [VC_BITS - 1:0] in_vc [0:IN_PORTS-1]; 
reg WEn [0:IN_PORTS-1];
   
reg [VC_BITS - 1:0] read_vc [0:IN_PORTS-1];
reg REn [0:IN_PORTS-1];
reg PeekEn [0:IN_PORTS-1];
  
wire [FLIT_WIDTH - 1:0] out_flit [0:IN_PORTS-1];
wire [VC_BITS - 1:0] out_vc [0:IN_PORTS-1];
wire valid [0:IN_PORTS-1];
  
wire [VC_PER_PORTS -1:0] empty [0:IN_PORTS-1];
wire [VC_PER_PORTS - 1:0] full [0:IN_PORTS-1]; 

reg [ADDRESS_BITS-1:0] t_in_address;
reg [(DATA_WIDTH * OFFSET)-1:0]  t_in_cache_line;
reg t_in_valid;
reg t_read, t_write, t_ack, t_reply2cache;
reg [ADDRESS_BITS-1:0] t_address;
reg [(DATA_WIDTH * OFFSET)-1:0]  t_mem_in_cache_line;    
  
wire [ID_BITS - 1: 0] source = CORE; 
reg  [ID_BITS - 1: 0] destination; 
reg  [ID_BITS - 1: 0] req_sender;
reg  [EXTRA - 1: 0]   sub_flow; 

reg [FLIT_WIDTH - 1:0] flit;
reg [FLOW_BITS - 1:0]  flow_ID; 
reg [TYPE_BITS - 1:0]  flit_type;
reg [VC_BITS - 1:0]    flit_vc;
reg [DATA_WIDTH - 1:0] flit_payload; 

reg  found; 
reg [VC_BITS - 1:0]    free_vc; 
reg [IN_PORTS-1 :0]    free_port; 
reg [1:0] ready_src;

reg  [(DATA_WIDTH * OFFSET)-1:0] buffers [0: (VC_PER_PORTS * IN_PORTS)-1];
reg  [DATA_WIDTH-1:0] addr_buffers [0: (VC_PER_PORTS * IN_PORTS)-1];
reg  [OFFSET-1:0] buf_counters [0: (VC_PER_PORTS * IN_PORTS)-1];
reg  cache_or_mem [0: (VC_PER_PORTS * IN_PORTS)-1];

reg [FLIT_WIDTH - 1:0] r_flit;
reg [FLOW_BITS - 1:0]  r_flow_ID; 
reg  [EXTRA - 1: 0]    r_sub_flow; 
reg [TYPE_BITS - 1:0]  r_flit_type;
reg [VC_BITS - 1:0]    r_flit_vc;
reg [DATA_WIDTH - 1:0] r_flit_payload;
reg [VC_BITS - 1:0]    vc;
reg [OFFSET-1:0]  counter;

reg [ADDRESS_BITS - 1:0] last_address; 
reg [ADDRESS_BITS - 1:0] last_req_address; 
reg [ADDRESS_BITS - 1:0] last_out_address; 
  
reg [2: 0] state; 
reg [2: 0] rec_state; 
reg [(PORT_BITS+VC_BITS) - 1:0] cur_vc_index; 
reg [PORT_BITS - 1:0] cur_port;
reg [VC_BITS- 1:0] cur_vc ;
wire [ADDRESS_BITS - REAL_ADDR_BITS -1 : 0]  zeros = 0; 
wire req_change = ((reset == 1) | (network_req == 0))? 1 : 
				  (state == REQ_ST)? 0 : req_change;
wire  [(DATA_WIDTH * OFFSET)-1:0]  mask = -1;
  
 assign ack = t_ack; 
 assign reply2cache  = t_reply2cache;
 assign read = t_read; 
 assign write = t_write; 
 assign address = t_address;
 assign mem_in_cache_line  =  t_mem_in_cache_line; 
 assign in_address = t_in_address; 
 assign in_cache_line  =  t_in_cache_line; 
 assign in_valid = t_in_valid;
 
integer index;     
genvar j;
 generate
		for (j=0; j < IN_PORTS; j=j+1) begin : IN_CORE_PORTS
			buffer_port #(FLIT_WIDTH, VC_BITS, VC_DEPTH_BITS) INP (
			  clock, reset, 
			  WEn[j], in_flit[j], in_vc [j], 
			  REn[j], PeekEn[j], read_vc[j],  
			  out_flit[j], out_vc[j], valid[j], 
			  empty[j], full[j]
			);
	    end
 endgenerate 
 generate
      	for (j = 0; j < OUT_PORTS; j = j+1) begin : OUT_CPORTS
      	 	assign t_to_core_flit[j]    = to_core_flit[((FLIT_WIDTH *(j + 1))-1) -: FLIT_WIDTH];
      	 	assign t_to_core_empty[j]   = to_core_empty[((VC_PER_PORTS *(j + 1))-1) -: VC_PER_PORTS]; 
      	 	assign t_to_core_full[j]    = to_core_full[((VC_PER_PORTS *(j + 1))-1) -: VC_PER_PORTS]; 
      	end  
      	assign t_v_to_core = v_to_core; 
      	
      	for (j = 0; j < IN_PORTS; j = j+1) begin : IN_CPORTS
			   assign from_core_flit [(((j + 1) *(FLIT_WIDTH))-1) -: FLIT_WIDTH]  = t_from_core_flit[j];
			   assign from_core_empty[(((j + 1) *(VC_PER_PORTS))-1) -: VC_PER_PORTS] = t_from_core_empty[j];
       	 assign from_core_full [(((j + 1) *(VC_PER_PORTS))-1) -: VC_PER_PORTS] = t_from_core_full[j];
      	end 	
      	assign v_from_core = t_v_from_core;
      	
      	for (j = 0; j < IN_PORTS; j = j+1) begin : EF_CPORTS
      			assign t_from_core_empty[j] = empty[j]; 
      			assign t_from_core_full[j]  = full[j]; 
      	end   	
 endgenerate 

always @ (posedge clock) begin
      if (reset) begin 
         	// Resetting 
            for(index= 0; index < IN_PORTS; index = index+1) begin
              	 WEn[index]           = 0;
              	 PeekEn[index]        = 0;        	 
            end
      end
      else begin    
         	// read in input ports 
            for(index= 0; index < IN_PORTS; index = index+1) begin
            	   WEn[index]         = (t_v_to_core[index]) ? 1 : 0; 
              	 if(t_v_to_core[index]) begin
              		  in_flit[index]     = t_to_core_flit[index];
              		  in_vc [index]      = t_to_core_flit[index] >> DATA_WIDTH; 
              	 end
            end 
      end
end 

  always @ (posedge clock) begin
    if(reset == 1) begin 
      state <= 	IDLE; 
      t_ack <= 0; 
      last_address <= 0;
      last_req_address <= 0;
      last_out_address <= 0; 
      destination      <= 0; 
      for(index= 0; index < IN_PORTS; index = index+1) begin
          t_from_core_flit[index]  <= 0;
	        t_v_from_core[index]    <= 0; 
      end
  end 
  else begin
     case (state)
      	IDLE: begin
      	     t_from_core_flit[free_port]  <= 0;
	         t_v_from_core[free_port]     <= 0;
	         t_ack                        <= 0;
	         t_reply2cache                <= 0;
      	    if( mem_valid || network_req || out_valid) begin
      	       found     <= 0; 
      	       free_vc   <= 0; 
      	       free_port <= 0; 
      	       state     <= SEARCH_ST; 
      	    end
      	    else state   <= IDLE;
      	end

	      SEARCH_ST: begin
            for(index= 0; index < IN_PORTS; index = index+1) begin
              	 	if((t_to_core_empty [index] > 0) && ~found) begin
              			   free_vc     <= t_to_core_empty [index] >> 1;
              			   free_port   <= index;
              			   found       <= 1; 
            		    end 
            end
            if(found)  state     <= DISPATCH;
            else state           <= SEARCH_ST;
      	end
      	
      	DISPATCH: begin
      	       if(mem_valid) state   <=  ((last_address == address) && (destination == req_sender) && (address > 0))? DISPATCH : READY_ST;
      		   else if(out_valid) state   <= (last_out_address == out_address)? DISPATCH : VAL_ST;
      		       else if(network_req) state   <= ((last_req_address == network_req_address) && (req_change == 0))? DISPATCH : REQ_ST;
      	     			else state   <= IDLE;         	     		       
      	end
    		
      	READY_ST: begin
      	 	counter        <= 0;	
      	 	last_address   <= address; 
      	 	destination    <= req_sender; 
      	 	sub_flow       <= 0;
         	flit_type      <= HEAD;
         	flit_vc        <= free_vc; 
         	flit_payload   <= address; 
         	state          <= SENT;    	         
      	end 
      	
      	REQ_ST: begin
      	 	counter        <= 0;
      	 	last_req_address   <= network_req_address;
      	 	destination    <= (network_req_address >> REAL_ADDR_BITS);
      	 	sub_flow       <= 0;
         	flit_type      <= ALL;
         	flit_vc        <= free_vc;
         	flit_payload   <= network_req_address;  
         	state          <= SENT;      	         
      	end
      	
      	VAL_ST: begin
      	 	counter        <= 0; 	 	
      	 	last_out_address   <= out_address;
      	 	destination    <= (out_address >> REAL_ADDR_BITS);
      	 	sub_flow       <= 1;
         	flit_type      <= HEAD;
         	flit_vc        <= free_vc;
         	flit_payload   <= out_address;   
         	state          <= SENT;    	         
      	end
      	                                       
        FILL_ST: begin
        	t_v_from_core[free_port]     <= 0;
            if(counter < OFFSET) begin
    	      counter        <= counter + 1; 
    	      destination    <= (sub_flow == 1)? (out_address >> REAL_ADDR_BITS): req_sender;
       	      flit_type      <= (counter == (OFFSET-1))? TAIL : BODY;
       	      flit_payload   <= (sub_flow == 1)? out_cache_line[(((counter + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH] 
     	                                        : mem_out_cache_line[(((counter + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH];   
         	  state          <= SENT;      
    	    end
    		else begin
    	      t_ack          <= (sub_flow == 1)? 0 : 1; 
    	      t_reply2cache  <= (sub_flow == 1)? 1 : 0; 
    		  state          <= IDLE; 
    		end   
        end  

      	SENT: begin
	        t_from_core_flit[free_port]  <= {source, destination, sub_flow, flit_type, flit_vc, flit_payload};
	        t_v_from_core[free_port]     <= 1;  
	        state                        <= (flit_type == ALL)? IDLE : FILL_ST;    	         
      	end 
      	
      	default:    state   <= IDLE;
      endcase 
   end   
end          

  always @ (posedge clock) begin
    if(reset == 1) begin 
      rec_state <= 	R_IDLE; 
      cur_vc_index <= 0; 
      cur_port <= 0; 
      cur_vc <= 0; 
      
      t_read    <= 0; 
      t_write   <= 0;
      t_address <= 0;
      req_sender <= 0; 
      t_mem_in_cache_line <= 0; 
      
      t_in_address     <= 0; 
      t_in_cache_line  <= 0;
      t_in_valid       <= 0;
		
		 for(index= 0; index < IN_PORTS; index = index+1) begin
          REn[index]           <= 0;       	 
       end
    end 
    else begin
      case (rec_state)
      	   R_IDLE: begin
			   cur_vc_index     <= (cur_vc_index == ((VC_PER_PORTS * IN_PORTS) -1))? 0 : cur_vc_index + 1;
			   cur_port         <= (cur_vc_index == ((VC_PER_PORTS * IN_PORTS) -1))? 0 : ((cur_vc_index + 1) >> VC_BITS);
 			   cur_vc           <= (cur_vc_index == ((VC_PER_PORTS * IN_PORTS) -1))? 0 : cur_vc_index + 1; 
			   rec_state        <= CHECK_VC;       
			   t_read           <= 0; 
               t_write          <= 0; 
               t_in_valid       <= 0; 		
      	   end

      	   CHECK_VC: begin	     
        	     if(empty[cur_port][cur_vc]== 0) begin
        	         REn[cur_port]      <= 1;  
				     read_vc[cur_port]  <= cur_vc; 
				     rec_state          <= READ_DELAY; 
        	     end   
      	       else rec_state         <= R_IDLE;
      	   end 
      	   
      	   READ_DELAY: begin	     
				          rec_state          <= GET_FLIT; 
				          REn[cur_port]      <= 0;
      	   end 
      	   
          GET_FLIT: begin
              if(valid[cur_port]==0) 
               	  $display("ERROR!!! READ ERROR MEM PROBLEM!!!!"); 
              else begin
				 	        r_flit			<= out_flit[cur_port]; 
				 	        r_flow_ID       <= out_flit[cur_port] >> (TYPE_BITS + VC_BITS + DATA_WIDTH);  
				 	        r_sub_flow      <= out_flit[cur_port] >> (TYPE_BITS + VC_BITS + DATA_WIDTH);  
				 	        r_flit_type     <= out_flit[cur_port] >> (VC_BITS + DATA_WIDTH); 
				 	        r_flit_vc       <= out_flit[cur_port] >> (DATA_WIDTH); 
			 	 	        r_flit_payload  <= out_flit[cur_port];  
              end
              rec_state           <=  PROCESS;
          end
 
          PROCESS: begin
              if(r_flit_type == ALL) begin
                 if(ready == 1) begin
                  	t_read         <= 1; 
                  	t_address      <= {zeros, r_flit_payload[REAL_ADDR_BITS -1 : 0]};
                  	req_sender     <= r_flit[(FLIT_WIDTH - 1) -: ID_BITS]; 
                  	rec_state      <= R_IDLE;   
                 end
                 else rec_state    <=  PROCESS; // loop
              end
              else begin
                  if (r_flit_type == HEAD)begin
                   	  buffers[cur_vc_index]        <= 0; 
					  addr_buffers[cur_vc_index]   <= r_flit_payload;
					  buf_counters[cur_vc_index]   <= 0;
					  cache_or_mem[cur_vc_index]   <= r_sub_flow;
					  rec_state                <= CHECK_VC;
               	   end
               	   else begin
             	        if(buf_counters[cur_vc_index] <(OFFSET-1)) begin
              	      	   buf_counters[cur_vc_index]   <= buf_counters[cur_vc_index] + 1;
                 	       buffers[cur_vc_index]<= ((buffers[cur_vc_index] & 
												   ((mask <<(((buf_counters[cur_vc_index]) + 1)* DATA_WIDTH)) ^ 
												   (mask >> ((OFFSET - (buf_counters[cur_vc_index]))* DATA_WIDTH)))) | 
									               (r_flit_payload <<((buf_counters[cur_vc_index])* DATA_WIDTH))); 
                 	       rec_state                <= CHECK_VC;
               	       end
               	       else begin
                 	        if(buf_counters[cur_vc_index] ==(OFFSET-1)) begin
                  	      	   buf_counters[cur_vc_index]   <= buf_counters[cur_vc_index] + 1;
										 buffers[cur_vc_index]<= ((buffers[cur_vc_index] & 
												  ((mask <<(((buf_counters[cur_vc_index]) + 1)* DATA_WIDTH)) ^ 
												  (mask >> ((OFFSET - (buf_counters[cur_vc_index]))* DATA_WIDTH)))) | 
									              (r_flit_payload <<((buf_counters[cur_vc_index])* DATA_WIDTH))); 
                     	        rec_state            <=  R_LAST;
                   	       end 
                   	       else  rec_state           <=  PROCESS;
               	       end            	       
   
                 	  end
                end
          end
    
          R_LAST: begin
             	  if(cache_or_mem[cur_vc_index] == 1) begin
             	  		if(ready == 1) begin
             	             t_address           <= {zeros, addr_buffers[cur_vc_index][REAL_ADDR_BITS -1 : 0]};
               	      	     t_mem_in_cache_line <=  buffers[cur_vc_index]; 
               	      	     t_write             <= 1; 
               	      	     rec_state           <= R_IDLE;
               	      	end
               	      	else rec_state           <= R_LAST; // loop
               	  end
            	  else begin 
            	      	t_in_address     <=  addr_buffers[cur_vc_index] | (r_flow_ID[(FLOW_BITS -1) -: ID_BITS] << REAL_ADDR_BITS); 
            	      	t_in_cache_line  <=  buffers[cur_vc_index]; 
                 	    t_in_valid       <=  1; 
                 	    rec_state        <=  R_IDLE;
            	  end
          end
                                            	
      	default:    rec_state   <= R_IDLE;
      endcase 
   end 
end
  
/*
 always @ (posedge clock) begin 
	$display ("-----------------------------Packetizer_Lite Module %d-------------------------------", CORE);  
	$display ("S State [%d]\t\t| R State [%d]", state, rec_state);

	$display ("free_vc[%b]\t\t| free_port  [%b]\t| found [%b]",free_vc, free_port , found);
	$display ("counter[%d]\t\t| destination [%b]\t| sub_flow [%b]",counter, destination , sub_flow );
	$display ("flit_type[%d]\t\t| flit_vc [%b]\t\t| flit_payload [%h]",flit_type, flit_vc , flit_payload);
	$display ("Reset [%b]\t\t| Read [%b]\t\t| Write [%b]\t\t| Ack [%b]",reset, read, write, ack);

	$display ("Last Address [%h]",last_address);
	$display ("Network_req [%b]", network_req); 
	$display ("Last_req_address [%h]", last_req_address); 
	$display ("Network_req_Address[%h] ", network_req_address);
	
	$display ("cur_vc_index [%b]\t| cur_port [%b]\t\t| cur_vc [%b]\t\t| empty [%b]",cur_vc_index, cur_port, cur_vc, empty[cur_port][cur_vc]);
	$display ("REn_cur_port [%b]\t| out_flit [%h]", REn[cur_port], out_flit[cur_port]);
	$display ("B_Counter[%d]\t| R_flow_ID [%b]| R_Sub_flow [%b]", buf_counters[cur_vc_index], r_flow_ID , r_sub_flow );
	$display ("R-Flit_type[%d]\t| R-Flit_vc [%b]\t\t| R-Flit_payload [%h]",r_flit_type, r_flit_vc , r_flit_payload);
	
	$display ("Out_valid [%b]", out_valid); 
	$display ("Out_address [%h] ", out_address);
	$display ("Out_Cache_line [%h] ", out_cache_line);
	
	$display ("In_valid [%b]", in_valid); 
	$display ("In_address [%h] ", in_address);
	$display ("In_Cache_line [%h] ", in_cache_line);
	
	$display ("Address [%h] ", address);
	$display ("Mem_in_Cache_line [%h] ", mem_in_cache_line);
	$display ("Ready [%b]\t\t| Mem_valid [%b]", ready, mem_valid); 
	$display ("Mem_out_Cache_line [%h] ", mem_out_cache_line); 

	$display ("Request Sender [%b] ", req_sender); 
	$display ("From Core Flit [%h] ", from_core_flit);
	$display ("From Core Valid [%b] ", v_from_core);
	$display ("From Core Empty [%b]\t| From Core Full [%b] ", from_core_empty, from_core_full);
	 
	$display ("To Core Flit [%h] ", to_core_flit);
	$display ("To Core Valid [%b] ", v_to_core);
	$display ("To Core Empty [%b]\t| To Core Full [%b] ", to_core_empty, to_core_full);
	$display ("Port [%d] Buf_Count [%d] Buf_Addr [%h] Buffer [%h] ", cur_port, buf_counters[cur_vc_index], addr_buffers[cur_vc_index], buffers[cur_vc_index]);
	$display ("-----------------------------------------------------------------------------------");
 end 	
//*/
               
endmodule
  
