 /** @module : Dir_cache_splitter
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
 
 module cache_splitter #(parameter CORE = 0, DATA_WIDTH = 32, INDEX_BITS = 6, 
                     OFFSET_BITS = 3, ADDRESS_BITS = 20, CACHE_ID_BITS = 5, 
					 ID_BITS = 4, MSG_BITS = 4)(
                     clock, reset,

					 cache2mem_msg, 
				     cache2mem_address, cache2mem_data,
					 
				     mem2cache_msg, 
				     mem2cache_address, mem2cache_data, 
					 
					 local2cache_msg, 
				     local2cache_address, local2cache_data,

					 cache2local_msg, 
				     cache2local_address, cache2local_data,

					 network2cache_msg, 
				     network2cache_address, network2cache_data,
					 
					 cache2network_msg, 
				     cache2network_address, cache2network_data
 ); 

input clock, reset; 

input [(MSG_BITS * 2)-1:0]      cache2mem_msg; 
input [(ADDRESS_BITS * 2)-1:0]  cache2mem_address; 
input [(DATA_WIDTH * 2)-1:0]    cache2mem_data;

output [(MSG_BITS * 2)-1:0]     mem2cache_msg; 
output [(ADDRESS_BITS * 2)-1:0] mem2cache_address; 
output [(DATA_WIDTH * 2)-1:0]   mem2cache_data;

input [MSG_BITS-1:0]            local2cache_msg; 
input [ADDRESS_BITS-1:0]        local2cache_address; 
input [DATA_WIDTH-1:0]          local2cache_data;

output [MSG_BITS-1:0]           cache2local_msg; 
output [ADDRESS_BITS-1:0]       cache2local_address; 
output [DATA_WIDTH-1:0]         cache2local_data;

input [MSG_BITS-1:0]            network2cache_msg; 
input [ADDRESS_BITS-1:0]        network2cache_address; 
input [DATA_WIDTH-1:0]          network2cache_data;

output [MSG_BITS-1:0]           cache2network_msg; 
output [ADDRESS_BITS-1:0]       cache2network_address; 
output [DATA_WIDTH-1:0]         cache2network_data;

wire [MSG_BITS-1:0]      i_cache2mem_msg       = cache2mem_msg[MSG_BITS-1: 0]; 
wire [ADDRESS_BITS-1:0]  i_cache2mem_address   = cache2mem_address[ADDRESS_BITS-1 : 0]; 
wire [DATA_WIDTH-1:0]    i_cache2mem_data      = cache2mem_data[DATA_WIDTH-1 : 0]; 

wire [MSG_BITS-1:0]      d_cache2mem_msg       = cache2mem_msg[((2*MSG_BITS)-1) -: MSG_BITS]; 
wire [ADDRESS_BITS-1:0]  d_cache2mem_address   = cache2mem_address[((2*ADDRESS_BITS)-1) -: ADDRESS_BITS]; 
wire [DATA_WIDTH-1:0]    d_cache2mem_data      = cache2mem_data[((2*DATA_WIDTH)-1) -: DATA_WIDTH];

localparam INFO_WIDTH = MSG_BITS + ADDRESS_BITS + DATA_WIDTH; 
localparam NO_MSG = 0;
localparam M_REQ = 1, INV_REQ = 2, M_WB_REQ = 3, SHARE_REQ = 4, WB_INV_REQ = 5; //memory requests
localparam M_READY = 6, MEM_SENT = 7,  W_RESP = 8, EX_RESP = 9, SH_RESP = 10;   //memory responses 
localparam R_REQ = 1, W_REQ = 2, C_WB_REQ = 3;                                  //cache requests
localparam C_READY = 4, C_SENT = 5, C_RECV = 6, INV_RESP = 7, WB_RESP = 8, WB_INV_RESP = 9;
//cache responses 

wire [INFO_WIDTH-1:0] i_cache2mem_info = {i_cache2mem_msg, i_cache2mem_address, i_cache2mem_data}; 
wire write_i_info = (i_cache2mem_msg != NO_MSG); 

wire [INFO_WIDTH-1:0] d_cache2mem_info = {d_cache2mem_msg, d_cache2mem_address, d_cache2mem_data}; 
wire write_d_info = (d_cache2mem_msg != NO_MSG);
 
wire valid;
wire [INFO_WIDTH - 1:0] read_data; 
wire [ID_BITS -1: 0]    core_id = CORE; 
 
wire [MSG_BITS-1:0]      merger_msg     = reset? 0 : valid? read_data[(INFO_WIDTH-1) -: MSG_BITS]: 0; 
wire [ADDRESS_BITS-1:0]  merger_address = reset? 0 : valid? read_data[(INFO_WIDTH-MSG_BITS-1) -: ADDRESS_BITS] : 0;  
wire [DATA_WIDTH-1:0]    merger_data    = reset? 0 : valid? read_data[DATA_WIDTH-1 : 0]: 0; 
wire [ID_BITS -1: 0]     dest = merger_address[ADDRESS_BITS-1 -:ID_BITS]; 

wire [INFO_WIDTH-1:0] network_info = {network2cache_msg, network2cache_address, network2cache_data}; 
wire write_n_info = (network2cache_msg != NO_MSG); 

wire [INFO_WIDTH-1:0] local2cache_info = {local2cache_msg, local2cache_address, local2cache_data}; 
wire write_l_info = (local2cache_msg != NO_MSG);
 
wire m_valid;
wire [INFO_WIDTH - 1:0] m_read_data; 
wire out_mem; 

wire [MSG_BITS-1:0]    m_merger_msg     = reset? 0 : m_valid? m_read_data[INFO_WIDTH-1 -: MSG_BITS]: NO_MSG; 
wire [ADDRESS_BITS-1:0]m_merger_address = reset? 0 : m_valid? m_read_data[(INFO_WIDTH -MSG_BITS-1) -: ADDRESS_BITS] : 0;  
wire [DATA_WIDTH-1:0]  m_merger_data    = reset? 0 : m_valid? m_read_data[DATA_WIDTH-1 : 0]: 0; 

reg  [CACHE_ID_BITS -1: 0] l_t_mem_dest, n_t_mem_dest; 
wire [CACHE_ID_BITS -1: 0] l_mem_dest = ((out_mem == 0) & ((m_merger_msg == INV_REQ) | 
							   (m_merger_msg == WB_INV_REQ) | (m_merger_msg == SHARE_REQ)| 
							   (m_merger_msg == M_WB_REQ) | (m_merger_msg == W_RESP) | 
							   (m_merger_msg == EX_RESP) | (m_merger_msg == SH_RESP) | 
							   (m_merger_msg == M_READY)))? m_merger_data : l_t_mem_dest; 
							   
wire [CACHE_ID_BITS -1: 0] n_mem_dest = ((out_mem == 1) & ((m_merger_msg == INV_REQ) | 
							   (m_merger_msg == WB_INV_REQ) | (m_merger_msg == SHARE_REQ)| 
							   (m_merger_msg == M_WB_REQ) | (m_merger_msg == W_RESP) | 
							   (m_merger_msg == EX_RESP) | (m_merger_msg == SH_RESP) | 
							   (m_merger_msg == M_READY)))? m_merger_data : n_t_mem_dest; 
				 
wire [MSG_BITS-1:0]     pad_msg     = 0; 
wire [ADDRESS_BITS-1:0] pad_address = 0; 
wire [DATA_WIDTH-1:0]   pad_data    = 0;

assign cache2local_msg      = (dest == core_id)? merger_msg : NO_MSG; 
assign cache2local_address  = (dest == core_id)? merger_address : 0; 
assign cache2local_data     = (dest == core_id)? merger_data : 0; 

assign cache2network_msg    = (dest != core_id)? merger_msg : NO_MSG; 
assign cache2network_address= (dest != core_id)? merger_address : 0; 
assign cache2network_data   = (dest != core_id)? merger_data : 0;
							  
assign mem2cache_msg     = reset? NO_MSG : (out_mem == 0)? 
				           ((l_mem_dest[0] == 0)? {pad_msg, m_merger_msg}:{m_merger_msg, pad_msg}): 
						   ((n_mem_dest[0] == 0)? {pad_msg, m_merger_msg}:{m_merger_msg, pad_msg}); 
assign mem2cache_address = reset? 0 : (out_mem == 0)? 
				           ((l_mem_dest[0] == 0)? {pad_address, m_merger_address}:{m_merger_address, pad_address}): 
						   ((n_mem_dest[0] == 0)? {pad_address, m_merger_address}:{m_merger_address, pad_address}); 
assign mem2cache_data    = reset? 0 : (out_mem == 0)? 
				           ((l_mem_dest[0] == 0)? {pad_data, m_merger_data}:{m_merger_data, pad_data}): 
						   ((n_mem_dest[0] == 0)? {pad_data, m_merger_data}:{m_merger_data, pad_data}); 

merger #(INFO_WIDTH, 2, 3, 0) CACM (
  .clock(clock), 
  .reset(reset), 
  .write({write_d_info, write_i_info}), 
  .write_data({d_cache2mem_info, i_cache2mem_info}), 

  .read(1'b1), 
  .read_data(read_data), 
  .out_fifo(), 
  .valid(valid), 
  
  .empty(), 
  .full()
); 

merger #(INFO_WIDTH, 2, 3, 0) MEMM (
  .clock(clock), 
  .reset(reset), 
  .write({write_n_info, write_l_info}), 
  .write_data({network_info, local2cache_info}), 

  .read(1'b1), 
  .read_data(m_read_data), 
  .out_fifo(out_mem), 
  .valid(m_valid), 
  
  .empty(), 
  .full()
); 
 
always @ (posedge clock) begin
	l_t_mem_dest     <= reset? 0: l_mem_dest; 
	n_t_mem_dest     <= reset? 0: n_mem_dest; 
end 

/*
 always @ (posedge clock) begin   
	$display ("--------------------------------Cache splitter core %d ----------------------------------", CORE);  
	$display ("cache2mem_msg [%h]\t| cache2mem_address [%h]| cache2mem_data [%h]", cache2mem_msg, cache2mem_address,cache2mem_data);
	$display ("mem2cache_msg [%h]\t| mem2cache_address [%h]| mem2cache_data [%h]", mem2cache_msg, mem2cache_address,mem2cache_data);
	$display ("m_merger_msg [%d]\t| m_merger_address [%h]     | m_merger_data [%h]", m_merger_msg, m_merger_address, m_merger_data);
	$display ("l_mem_dest [%d]\t| n_mem_dest [%d]", l_mem_dest, n_mem_dest);
	$display ("icache2mem_msg [%h]\t| icache2mem_address [%h]  | icache2mem_data [%h]", i_cache2mem_msg, i_cache2mem_address,i_cache2mem_data);
	$display ("dcache2mem_msg [%h]\t| dcache2mem_address [%h]  | dcache2mem_data [%h]", d_cache2mem_msg, d_cache2mem_address,d_cache2mem_data);
	$display ("cache2local_msg [%d]\t| cache2local_address [%h] | cache2local_data [%h]", cache2local_msg, cache2local_address,cache2local_data); 
	$display ("local2cache_msg [%d]\t| local2cache_address [%h] | local2cache_data [%h]", local2cache_msg, local2cache_address,local2cache_data);
	$display ("cache2network_msg [%d]| cache2network_address [%h]| cache2network_data [%h]", cache2network_msg, cache2network_address,cache2network_data);
	$display ("network2cache_msg [%d]| network2cache_address [%h]| network2cache_data [%h]", network2cache_msg, network2cache_address,network2cache_data);
	$display ("-----------------------------------------------------------------------------------");
 end
// /*/
		
endmodule 