/** @module : Cache_wrapper
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
module cache_wrapper #(parameter CORE = 0, DATA_WIDTH = 32, INDEX_BITS = 6, 
                     OFFSET_BITS = 3, ADDRESS_BITS = 20, MSG_BITS = 4) (
				     clock, reset,  
				     //Core2Cache
				     read, write, address, in_data, 
				     out_addr, out_data, valid, ready,
				     
				     //Cache2Memory
				     mem2cache_msg, 
				     mem2cache_address, mem2cache_data, 
				     
				     cache2mem_msg, 
				     cache2mem_address, cache2mem_data, 
					 // performance reporting
					 report
);

localparam OFFSET   = 1<<OFFSET_BITS; 
localparam BLOCK_BITS  = DATA_WIDTH * OFFSET; 

input clock, reset; 
//Core2Cache
input read, write;
input [ADDRESS_BITS-1:0] address;
input [DATA_WIDTH-1:0] in_data;
output valid, ready;
output[ADDRESS_BITS-1:0] out_addr;
output[DATA_WIDTH-1:0] out_data;

//Cache2Memory
input [MSG_BITS-1:0]      mem2cache_msg; 
input [ADDRESS_BITS-1:0]  mem2cache_address; 
input [DATA_WIDTH-1:0]    mem2cache_data;
 				     
output [MSG_BITS-1:0]     cache2mem_msg; 
output [ADDRESS_BITS-1:0] cache2mem_address; 
output [DATA_WIDTH-1:0]   cache2mem_data; 
input  report;
  
reg [MSG_BITS-1:0]    t_cache2mem_msg; 
reg [ADDRESS_BITS-1:0]t_cache2mem_address; 
reg [DATA_WIDTH-1:0]  t_cache2mem_data;

wire mem2cache_share;
wire mem2cache_req_share;
wire mem2cache_exclusive;
wire mem2cache_modify;
wire mem2cache_invalidate;
wire mem2cache_writeback;
wire [ADDRESS_BITS-1:0] mem2cache_addr;
wire [BLOCK_BITS-1:0]   mem2cache_line;
				     
wire cache2mem_read;
wire cache2mem_write;
wire cache2mem_invalid;
wire cache2mem_share; 
wire cache2mem_writeback;
wire [ADDRESS_BITS-1:0] cache2mem_addr;
wire [BLOCK_BITS-1:0]   cache2mem_line; 

wire [MSG_BITS-1:0]      in_msg; 
wire [ADDRESS_BITS-1:0]  in_address; 
wire [BLOCK_BITS-1:0]    in_data_line;
 				     
wire [MSG_BITS-1:0]     out_msg; 
wire [ADDRESS_BITS-1:0] out_address; 
wire [BLOCK_BITS-1:0]   out_data_line;

dir_direct_mapped_cache #(CORE, DATA_WIDTH, INDEX_BITS, OFFSET_BITS, ADDRESS_BITS) CC_CACHE (
					.clock(clock), 
					.reset(reset),  
					//Core2Cache
					.read(read), 
					.write(write), 
					.address(address), 
					.in_data(in_data), 
					.out_addr(out_addr), 
					.out_data(out_data), 
					.valid(valid), 
					.ready(ready),
				     
					//Cache2MemorySubSystem
					.mem2cache_share(mem2cache_share), 
					.mem2cache_req_share(mem2cache_req_share),
					.mem2cache_exclusive(mem2cache_exclusive),
					.mem2cache_modify(mem2cache_modify),
					.mem2cache_invalidate(mem2cache_invalidate), 
					.mem2cache_writeback(mem2cache_writeback),
					.mem2cache_address(mem2cache_addr), 
					.mem2cache_line(mem2cache_line), 

					.cache2mem_read(cache2mem_read), 
					.cache2mem_write(cache2mem_write), 
					.cache2mem_invalid(cache2mem_invalid), 
					.cache2mem_share(cache2mem_share), 
					.cache2mem_writeback(cache2mem_writeback), 
					.cache2mem_address(cache2mem_addr), 
					.cache2mem_line(cache2mem_line), 
					.report(report)
);
					 
cache_translator #(CORE, DATA_WIDTH, OFFSET_BITS, ADDRESS_BITS,  MSG_BITS) C_DIR (
                     .clock(clock), 
					 .reset(reset),
                     
					 .read(cache2mem_read), 
					 .write(cache2mem_write), 
					 .flush(cache2mem_writeback), 
					 .invalid(cache2mem_invalid), 
					 .share(cache2mem_share), 
					 .address(cache2mem_addr), 
					 .in_line(cache2mem_line),
					 
				     .invalidate(mem2cache_invalidate), 
					 .sharing(mem2cache_share), 
					 .sharing_req(mem2cache_req_share), 
					 .write_back(mem2cache_writeback), 
					 .exclusive(mem2cache_exclusive), 
					 .modify(mem2cache_modify), 
					 .out_addr(mem2cache_addr), 
					 .out_line(mem2cache_line),
					 
				     .in_msg(in_msg), 
					 .in_address(in_address), 
					 .in_data_line(in_data_line),
				     .out_msg(out_msg), 
					 .out_address(out_address),
					 .out_data_line(out_data_line)
 ); 
 
cache_interface #(CORE, DATA_WIDTH, OFFSET_BITS, ADDRESS_BITS,  MSG_BITS) C_Inter (
                     .clock(clock), 
					 .reset(reset),
				     
				     .in_msg(out_msg),  
					 .in_address(out_address),  
					 .in_data_line(out_data_line),  
				     .out_msg(in_msg),  
					 .out_address(in_address),  
					 .out_data_line(in_data_line),
				     
				     .mem2cache_msg(mem2cache_msg),  
				     .mem2cache_address(mem2cache_address),  
					 .mem2cache_data(mem2cache_data),  
				     
				     .cache2mem_msg(cache2mem_msg),  
				     .cache2mem_address(cache2mem_address),  
					 .cache2mem_data(cache2mem_data)
);						

/*
	always @ (posedge clock) begin   
		$display ("-------------------------- Cache Wrapper [%d]-------------------------------", CORE);   
		$display ("Reset [%b]\t\t| Read [%b]\t\t| Write [%b]\t\t",reset, read, write);
		$display ("New Addr [%h]\t| New In_data [%h]", address, in_data);
		$display ("Cache2mem_read[%b]\t| Cache2mem_write[%b]\t| Cache2mem_invalid[%b]", cache2mem_read, cache2mem_write, cache2mem_invalid); 		
		$display ("Cache2mem_share[%b]\t| Cache2mem_writeback[%b]", cache2mem_share, cache2mem_writeback); 		
		$display ("Cache2mem_address[%h]", cache2mem_addr); 
		$display ("Cache2mem_line[%h]", cache2mem_line); 	
		$display ("Mem2cache_share[%b]\t| Mem2cache_exclusive[%b]| Mem2cache_modify[%b]", mem2cache_share,  mem2cache_exclusive, mem2cache_modify);
		$display ("Mem2cache_invalidate[%b]|Mem2cache_writeback[%b]", mem2cache_invalidate, mem2cache_writeback); 		
		$display ("Mem2cache_address[%h]", mem2cache_address); 
		$display ("Mem2cache_line[%h]", mem2cache_line);   
		$display (""); 
		$display ("cache2mem_msg [%d]\t| cache2mem_address [%h]| cache2mem_data [%h]", cache2mem_msg, cache2mem_address,cache2mem_data);
		$display ("mem2cache_msg [%d]\t| mem2cache_address [%h]| mem2cache_data [%h]", mem2cache_msg, mem2cache_address,mem2cache_data);
		$display ("-----------------------------------------------------------------------------");
	end 	
//*/
endmodule

