/** @module : directory
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
 
 module directory #(parameter CORE = 0, ADDRESS_BITS = 20, MSG_BITS = 4, ID_BITS = 2, NUM_SHARERS_BITS = 2)(
                     clock, reset, 
					 resp_ready, req_ready, 
					 
				     in_msg, 
					 src_id,
				     in_address, 
				     
				     out_msg,
					 dest_id,
				     out_address
 ); 

input clock, reset; 
output resp_ready, req_ready; 

input [MSG_BITS-1:0]        in_msg; 
input [ID_BITS-1:0]         src_id;
input [ADDRESS_BITS-1:0]    in_address; 
 				     
output [MSG_BITS-1:0]       out_msg; 
output [ID_BITS-1:0]        dest_id;
output [ADDRESS_BITS-1:0]   out_address; 

localparam NO_MSG = 0;
localparam M_REQ = 1, INV_REQ = 2, M_WB_REQ = 3, SHARE_REQ = 4, WB_INV_REQ = 5; //memory requests
localparam M_READY = 6, MEM_SENT = 7,  W_RESP = 8, EX_RESP = 9, SH_RESP = 10;   //memory responses 
localparam R_REQ = 1, W_REQ = 2, C_WB_REQ = 3;                                  //cache requests
localparam C_READY = 4, C_SENT = 5, C_RECV = 6, INV_RESP = 7, WB_RESP = 8, WB_INV_RESP = 9;
//cache responses 

localparam FREE = 0, SHARED = 1, EXCLUSIVE = 2, MODIFIED = 3;

localparam IDLE = 0, READ = 1, WRITE = 2, READ_DIR = 3, WRITE_DIR = 4; 
localparam WAIT = 5, DONE_READ = 6, DONE_WRITE = 7; 
localparam RE_TRY = 8, UPDATE = 9, UPDATE_DONE = 10; 

localparam STATUS_BITS = 2; 
localparam NUM_SHARERS = (1 << NUM_SHARERS_BITS); 
localparam ENTRY_BITS = STATUS_BITS + NUM_SHARERS + (NUM_SHARERS * ID_BITS); 

reg read, write; 
reg [ADDRESS_BITS-1:0] address; 

wire [ENTRY_BITS-1:0] entry_in; 
wire [ENTRY_BITS-1:0] entry_out;

wire [STATUS_BITS-1:0] status; 
wire [NUM_SHARERS-1:0] v_sharers;
wire [ID_BITS-1:0]     sharer [0 : NUM_SHARERS-1]; 

reg t_req_ready;
reg [MSG_BITS-1:0]       t_out_msg; 
reg [ID_BITS-1:0]        t_dest_id;
reg [ADDRESS_BITS-1:0]   t_out_address; 

reg [3:0] state; 
reg [ID_BITS-1:0] pe_id;

reg [STATUS_BITS-1:0] t_status; 
reg [NUM_SHARERS-1:0] t_v_sharers;
reg [ID_BITS-1:0]     t_sharer [0 : NUM_SHARERS-1];
reg [ID_BITS-1:0]     cur_sharer;

reg [MSG_BITS-1:0]    cur_msg;

wire no_wait = (((t_out_msg == INV_REQ)|(t_out_msg == SHARE_REQ)|(t_out_msg == M_WB_REQ)) &
				(src_id == t_dest_id) & (in_address == t_out_address)); 

// To update other responses that are coming while waiting on another address				
reg side_read, prev_side_read, side_write; 
reg [ADDRESS_BITS-1:0] side_address, prev_side_address; 
reg [MSG_BITS-1:0]    side_msg, prev_side_msg;
reg [ID_BITS-1:0] side_pe_id, prev_side_pe_id;

wire [ENTRY_BITS-1:0] side_entry_in; 
wire [ENTRY_BITS-1:0] side_entry_out;

wire [STATUS_BITS-1:0] side_status; 
wire [NUM_SHARERS-1:0] side_v_sharers;
wire [ID_BITS-1:0]     side_sharer [0 : NUM_SHARERS-1]; 

reg [STATUS_BITS-1:0] side_t_status; 
reg [NUM_SHARERS-1:0] side_t_v_sharers;
reg [ID_BITS-1:0]     side_t_sharer [0 : NUM_SHARERS-1];

integer u; 
wire [NUM_SHARERS-1:0]  slot      = get_slot(v_sharers);
wire [NUM_SHARERS-1:0]  pos       = free_slot(v_sharers);

function  sharing;
	input [NUM_SHARERS-1:0] v_sharers;
	integer i; 
	reg sharer1, sharer2;
	begin
		sharer1 = 0; sharer2 = 0;
		for(i= 0; i < NUM_SHARERS; i = i+1) begin
			if (v_sharers[i]) begin 
				if (sharer1) begin sharer2 = 1; end 
				else begin sharer1 = 1; end 
			end 			
		end 
		sharing = sharer2;
	end
endfunction

function  [NUM_SHARERS-1:0] get_slot;
	input [NUM_SHARERS-1:0] v_sharers;
	integer i, index; 
	integer found;
	begin
		found = 0; index = 0; 
		for(i= 0; i < NUM_SHARERS; i = i+1) begin
			if ((v_sharers[i] == 0) & (found == 0))begin 
				found = 1; 
				index = i; 
			end 			
		end 
		get_slot = index;
	end
endfunction

function  [NUM_SHARERS-1:0] free_slot;
	input [NUM_SHARERS-1:0] v_sharers;
	integer i, index; 
	integer found;
	begin
		found = 0; index = 0; 
		for(i= 0; i < NUM_SHARERS; i = i+1) begin
			if ((v_sharers[i] == 1) & (found == 0))begin 
				if (sharer[i] != pe_id) found = 1; 
				index = i; 
			end 			
		end 
		free_slot = index;
	end
endfunction

two_port_BRAM #(ENTRY_BITS, ADDRESS_BITS) DIRECTORY(
		.clock(clock),
    	.readEnable0(read),
    	.readAddress0(address),
   	    .readData0(entry_out),
    	.writeEnable0(write),
    	.writeAddress0(address),
    	.writeData0(entry_in),
    	.readEnable1(side_read),
    	.readAddress1(side_address),
   	    .readData1(side_entry_out),
    	.writeEnable1(side_write),
    	.writeAddress1(side_address),
    	.writeData1(side_entry_in)
);
				
genvar i;
generate   
	assign status    = entry_out[(ENTRY_BITS-1) -: STATUS_BITS]; 
	assign v_sharers = entry_out[(ENTRY_BITS-STATUS_BITS-1) -: NUM_SHARERS]; 
	for (i = 0; i < NUM_SHARERS; i=i+1) begin : IOUT_ENT
		assign sharer[i] = entry_out[(((i + 1) *(ID_BITS))-1) -: ID_BITS];
	end 
		
	assign entry_in[(ENTRY_BITS-1) -: STATUS_BITS]             = t_status; 
	assign entry_in[(ENTRY_BITS-STATUS_BITS-1) -: NUM_SHARERS] = t_v_sharers; 
	for (i = 0; i < NUM_SHARERS; i=i+1) begin : IN_ENT
		assign entry_in[(((i + 1) *(ID_BITS))-1) -: ID_BITS]   = t_sharer[i];
	end
endgenerate 

generate   
	assign side_status    = side_entry_out[(ENTRY_BITS-1) -: STATUS_BITS]; 
	assign side_v_sharers = side_entry_out[(ENTRY_BITS-STATUS_BITS-1) -: NUM_SHARERS]; 
	for (i = 0; i < NUM_SHARERS; i=i+1) begin : S_IOUT_ENT
		assign side_sharer[i] = side_entry_out[(((i + 1) *(ID_BITS))-1) -: ID_BITS];
	end 
		
	assign side_entry_in[(ENTRY_BITS-1) -: STATUS_BITS]             = side_t_status; 
	assign side_entry_in[(ENTRY_BITS-STATUS_BITS-1) -: NUM_SHARERS] = side_t_v_sharers; 
	for (i = 0; i < NUM_SHARERS; i=i+1) begin : S_IN_ENT
		assign side_entry_in[(((i + 1) *(ID_BITS))-1) -: ID_BITS]   = side_t_sharer[i];
	end
endgenerate 

// To update other responses that are coming while waiting on another address	
always @ (posedge clock) begin
	if (reset) begin
		side_read           <= 0; 
		side_write          <= 0; 
		side_address        <= 0;  
		side_msg            <= 0;
		side_pe_id          <= 0;
		
		prev_side_read      <= 0; 
		prev_side_address   <= 0;
		prev_side_msg       <= 0;
		prev_side_pe_id     <= 0;
	end 
	else begin 
		side_read           <= (((in_msg == INV_RESP)|(in_msg == SH_RESP)|(in_msg == WB_RESP) |
							   (in_msg == WB_INV_RESP)) & (state == WAIT) & (in_address != address))? 1 : 0; 
		side_address        <= (((in_msg == INV_RESP)|(in_msg == SH_RESP)|(in_msg == WB_RESP) |
							   (in_msg == WB_INV_RESP)) & (state == WAIT) & (in_address != address))? in_address : 0; 
		side_msg            <= (((in_msg == INV_RESP)|(in_msg == SH_RESP)|(in_msg == WB_RESP) |
							   (in_msg == WB_INV_RESP)) & (state == WAIT) & (in_address != address))? in_msg : NO_MSG; 
		side_pe_id          <= src_id;
		
		prev_side_read      <= side_read; 
		prev_side_address   <= side_address;
		prev_side_msg       <= side_msg;
		prev_side_pe_id     <= side_pe_id;
		
			case (prev_side_msg)
				INV_RESP: begin
					for (u = 0; u < NUM_SHARERS; u=u+1) begin
						if ((side_sharer[u] == prev_side_pe_id) & side_v_sharers[u]) begin 
							side_t_status        <= (side_t_status == EXCLUSIVE)? FREE : sharing(side_v_sharers)? SHARED : FREE; 
							side_t_v_sharers[u]  <= 0; 
							side_t_sharer[u]	 <= 0;	
						end 
						else begin 
							side_t_sharer[u]    <=  side_sharer[u]; 
							side_t_v_sharers[u] <=  side_v_sharers[u]; 
						end 
					end 
					side_write <= 1;
				end
				SH_RESP: begin
					for (u = 0; u < NUM_SHARERS; u=u+1) begin
						if ((side_sharer[u] == prev_side_pe_id) & side_v_sharers[u]) begin 
							side_t_status        <= SHARED; 
							side_t_sharer[u]	 <= side_sharer[u];	
							side_t_v_sharers[u]  <= 1; 
						end 
						else begin 
							side_t_sharer[u]    <=  side_sharer[u]; 
							side_t_v_sharers[u] <=  side_v_sharers[u];
						end 
					end 
					side_write <= 1; 
				end
				WB_RESP: begin
					for (u = 0; u < NUM_SHARERS; u=u+1) begin
						if ((side_sharer[u] == prev_side_pe_id) & side_v_sharers[u]) begin 
							side_t_status        <= SHARED;
							side_t_sharer[u]	 <= side_sharer[u];							
							side_t_v_sharers[u]  <= 1; 
						end 
						else begin 
							side_t_sharer[u]    <=  side_sharer[u]; 
							side_t_v_sharers[u] <=  side_v_sharers[u];
						end 
					end 
					side_write <= 1; 
				end
				WB_INV_RESP: begin
					for (u = 0; u < NUM_SHARERS; u=u+1) begin
						if ((side_sharer[u] == prev_side_pe_id) & side_v_sharers[u]) begin 
							side_t_status        <= FREE;
							side_t_sharer[u]	 <= 0; 						
							side_t_v_sharers[u]  <= 0; 
						end 
						else begin 
							side_t_sharer[u]    <=  side_sharer[u]; 
							side_t_v_sharers[u] <=  side_v_sharers[u];
						end
					end 
					side_write <= 1;					
				end
				default: side_write <= 0; 
			endcase 	
	end
end 

always @ (posedge clock) begin
	if (reset) begin
		read           <= 0; 
		write          <= 0; 
		address        <= 0;  
		
		t_req_ready    <= 1;
		t_out_msg      <= NO_MSG; 
		t_dest_id      <= 0;
		t_out_address  <= 0;
		
		cur_msg        <= NO_MSG; 
		state          <= IDLE; 
	end 
	else begin
		case (state)
		IDLE: begin
			read           <= ((in_msg == R_REQ) | (in_msg == W_REQ) 
							  | (in_msg == WB_INV_RESP) | (in_msg == INV_RESP))? 1 : 0; 
			write          <= 0;
			pe_id          <= src_id; 
			address        <= in_address;
			
			t_req_ready    <= ((in_msg == R_REQ) | (in_msg == W_REQ) 
							  | (in_msg == WB_INV_RESP) | (in_msg == INV_RESP))? 0 : 1; 
			t_out_msg      <= NO_MSG; 
			t_dest_id      <= 0;
			t_out_address  <= 0;
			
			cur_msg        <= in_msg; 
			state          <= (in_msg == R_REQ)? READ : (in_msg == W_REQ)?  WRITE : 
			                  (in_msg == M_WB_REQ)? READ :
							  ((in_msg == WB_INV_RESP) | (in_msg == INV_RESP))? UPDATE : IDLE; 
		end

		RE_TRY: begin
			read           <= ((cur_msg == R_REQ) | (cur_msg == W_REQ))? 1 : 0; 
			write          <= 0;
			
			t_out_msg      <= NO_MSG; 
			t_dest_id      <= 0;
			t_out_address  <= 0;
			
			state          <= (cur_msg == R_REQ)? READ : (cur_msg == W_REQ)?  WRITE : 
			                  (cur_msg == M_WB_REQ)? READ : IDLE; 
		end
		
		READ: begin
			read        <= 0; 
			state       <= READ_DIR;
		end 

		READ_DIR: begin 			
			case (status)
				FREE: begin
					t_out_msg      <= EX_RESP; 
					t_dest_id      <= pe_id;
					t_out_address  <= address;
					
					t_status       <= EXCLUSIVE; 
					t_v_sharers    <= 1; 
					t_sharer[0]    <=  pe_id; 
					for (u = 1; u < NUM_SHARERS; u=u+1) begin
						t_sharer[u]    <=  sharer[u]; 
					end 
					
					state          <= DONE_READ;
				end

				SHARED: begin
					if(~v_sharers) begin 
						t_out_msg      <= SH_RESP; 
						t_dest_id      <= pe_id;
						t_out_address  <= address;
					
						t_status       <= SHARED; 
						for (u = 0; u < NUM_SHARERS; u=u+1) begin
							if (u == slot) begin 
								t_sharer[u]    <=  pe_id; 
								t_v_sharers[u] <= 1;
							end 
							else begin 
								t_sharer[u]    <=  sharer[u]; 
								t_v_sharers[u] <=  v_sharers[u];
							end
						end 
						
						state          <= DONE_READ;
					end 
					else begin 
						t_status       <= status;
						
						t_out_msg      <= INV_REQ; 
						t_dest_id      <= sharer[0];
						t_out_address  <= address;
						
						read           <= 1;
						state          <= WAIT;
					end  
				end 
				
				EXCLUSIVE: begin
					t_status       <= status;
					
					t_out_msg      <= SHARE_REQ; 
					t_dest_id      <= sharer[0];
					t_out_address  <= address;
					read           <= 1;
					state          <= WAIT;
				end
				
				MODIFIED: begin
					t_status       <= status;
					
					t_out_msg      <= M_WB_REQ; 
					t_dest_id      <= sharer[0];
					t_out_address  <= address;
					
					read           <= 1;
					state          <= WAIT;
				end
			endcase  
		end 
		
		DONE_READ: begin
			write       <= 1; 
			t_req_ready <= 1; 
			t_out_msg   <= NO_MSG;
			cur_msg     <= NO_MSG; 
			state       <= IDLE;
		end 
		
		WRITE: begin
			read        <= 0; 
			state       <= WRITE_DIR;
		end 

		WRITE_DIR: begin 			
			case (status)
				FREE: begin
					t_out_msg      <= W_RESP; 
					t_dest_id      <= pe_id;
					t_out_address  <= address;
					
					t_status       <= MODIFIED; 
					t_v_sharers    <= 1; 
					t_sharer[0]    <=  pe_id; 
					for (u = 1; u < NUM_SHARERS; u=u+1) begin
						t_sharer[u]    <= 0; 
					end 
					
					state          <= DONE_WRITE;
				end

				SHARED: begin
					if(sharer[pos] != pe_id) begin 
						t_out_msg      <= INV_REQ; 
						t_dest_id      <= sharer[pos];
						t_out_address  <= address;
					
						t_status       <= status; 
						read           <= 1;
						state          <= WAIT;
					end 
					else begin 
						t_out_msg      <= W_RESP; 
						t_dest_id      <= sharer[pos];
						t_out_address  <= address;
						
						t_status       <= MODIFIED; 
						t_v_sharers    <= 1;
						t_sharer[0]    <=  sharer[pos]; 
						for (u = 1; u < NUM_SHARERS; u=u+1) begin
							t_sharer[u]    <=  0; 
						end 
						state          <= DONE_WRITE;
					end  
				end 
				
				EXCLUSIVE: begin
					if(sharer[0] == pe_id) begin 
						t_out_msg      <= W_RESP; 
						t_dest_id      <= sharer[0];
						t_out_address  <= address;
						
						t_status       <= MODIFIED; 
						t_v_sharers    <= 1;
						t_sharer[0]    <=  sharer[0]; 
						for (u = 1; u < NUM_SHARERS; u=u+1) begin
							t_sharer[u]    <=  0; 
						end 
						state          <= DONE_WRITE;
					end 
					else begin 
						t_out_msg      <= INV_REQ; 
						t_dest_id      <= sharer[0];
						t_out_address  <= address;
						
						t_status       <= status; 
						read           <= 1;
						state          <= WAIT;
					end 
				end
				
				MODIFIED: begin
					t_out_msg      <= WB_INV_REQ; 
					t_dest_id      <= sharer[0];
					t_out_address  <= address;
					
					t_status       <= status;
					read           <= 1;
					state          <= WAIT;
				end
			endcase  
		end 
		
		DONE_WRITE: begin
			write       <= 1; 
			t_req_ready <= 1;
			t_out_msg   <= NO_MSG;
			cur_msg     <= NO_MSG; 
			state       <= IDLE;
		end 
		
		UPDATE: begin
			read        <= 0; 
			state       <= UPDATE_DONE;
		end 
		
		UPDATE_DONE: begin
			for (u = 0; u < NUM_SHARERS; u=u+1) begin
				if ((sharer[u] == pe_id) & v_sharers[u]) begin 
					t_status        <= sharing(v_sharers)? SHARED : FREE; 
					t_sharer[u]     <= 0; 
					t_v_sharers[u]  <= 0; 
				end 
				else begin 
					t_sharer[u]    <=  sharer[u]; 
					t_v_sharers[u] <=  v_sharers[u];
				end
			end 
			read        <= 0; 
			write       <= 1; 
			state       <= IDLE; 
		end
		
		WAIT: begin
			t_out_msg      <= NO_MSG; 
			if (in_address == address) begin 
				case (in_msg)
					INV_RESP: begin
						for (u = 0; u < NUM_SHARERS; u=u+1) begin
							if ((sharer[u] == src_id) & v_sharers[u]) begin 
								t_status        <= (status == EXCLUSIVE)? FREE : sharing(v_sharers)? SHARED : FREE; 
								t_sharer[u]     <= 0;
								t_v_sharers[u]  <= 0; 
							end 
							else begin 
								t_sharer[u]    <=  sharer[u]; 
								t_v_sharers[u] <=  v_sharers[u]; 
							end 
						end 
						write <= 1;
					end
					SH_RESP: begin
						for (u = 0; u < NUM_SHARERS; u=u+1) begin
							if ((sharer[u] == src_id) & v_sharers[u]) begin 
								t_status        <= SHARED; 
								t_v_sharers[u]  <= 1; 
							end 
							else begin 
								t_v_sharers[u] <=  v_sharers[u];
							end 
							t_sharer[u]    <=  sharer[u];
						end 
						write <= 1; 
					end
					WB_RESP: begin
						for (u = 0; u < NUM_SHARERS; u=u+1) begin
							if ((sharer[u] == src_id) & v_sharers[u]) begin 
								t_status        <= SHARED; 
								t_v_sharers[u]  <= 1; 
							end 
							else begin 
								t_v_sharers[u] <=  v_sharers[u];
							end 
							t_sharer[u]    <=  sharer[u];
						end 
						write <= 1; 
					end
					WB_INV_RESP: begin
						for (u = 0; u < NUM_SHARERS; u=u+1) begin
							if ((sharer[u] == src_id) & v_sharers[u]) begin 
								t_status        <= FREE; 
								t_sharer[u]     <= 0;
								t_v_sharers[u]  <= 0; 
							end 
							else begin 
								t_sharer[u]    <=  sharer[u]; 
								t_v_sharers[u] <=  v_sharers[u];
							end
						end 
						write <= 1; 
					end
				endcase 
			end 
			state      <= (((in_msg == INV_RESP)|(in_msg == SH_RESP)|(in_msg == WB_RESP)|
						  (in_msg == WB_INV_RESP)) & (in_address == address))? RE_TRY : WAIT;
		end 
		endcase    

	end 
end

//----------------------------------------------------
// Drive the outputs
//----------------------------------------------------
assign resp_ready  = (reset | t_req_ready | (state == WAIT));
assign req_ready   = t_req_ready;
assign out_msg     = t_out_msg; 
assign dest_id     = t_dest_id;
assign out_address = t_out_address ; 

/*
always @ (posedge clock) begin
	$display ("------------------------------- Directory [%d]------------------------------------------", CORE);
	$display ("State [%d]\t| Status [%d]\t| Rp_ready [%b]\t| Rq_ready [%b]", state, status, resp_ready, req_ready);
	$display ("Read [%b]\t| Write [%b]", read, write);
	$display ("Cur_msg [%d]\t| PE [%d]\t| Address [%h]", cur_msg, pe_id, address);
	$display ("Out entry[%h]| Status [%d]\t| Valid [%b]", entry_out, entry_out[(ENTRY_BITS-1) -: STATUS_BITS], entry_out[(ENTRY_BITS-STATUS_BITS-1) -: NUM_SHARERS]); 
	$display ("In entry [%h]| Status [%d]\t| Valid [%b]", entry_in, entry_in[(ENTRY_BITS-1) -: STATUS_BITS], entry_in[(ENTRY_BITS-STATUS_BITS-1) -: NUM_SHARERS]); 
    $display ("--- Side operations ---");
	$display ("Read [%b]\t| Write [%b]", side_read, side_write);
	$display ("Side_msg [%d]\t| PE [%d]\t| Address [%h]", side_msg, side_pe_id, side_address);
	$display ("Out entry[%h]| Status [%d]\t| Valid [%b]", side_entry_out, side_entry_out[(ENTRY_BITS-1) -: STATUS_BITS], side_entry_out[(ENTRY_BITS-STATUS_BITS-1) -: NUM_SHARERS]); 
	$display ("In entry [%h]| Status [%d]\t| Valid [%b]",side_entry_in, side_entry_in[(ENTRY_BITS-1) -: STATUS_BITS], side_entry_in[(ENTRY_BITS-STATUS_BITS-1) -: NUM_SHARERS]); 
	$display ("IN_MSG [%d]\t| Src [%d]\t| Address [%h]", in_msg, src_id, in_address);
	$display ("OUT_MSG [%d]\t| Dest [%d]\t| Address [%h]", out_msg, dest_id, out_address);
	$display ("------------------------------------------------------------------------------------");
end 
//*/

endmodule