/** @module : main_memory_translator
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
 
module main_memory_translator #(parameter CORE = 0, DATA_WIDTH = 32, ADDRESS_BITS = 20, 
					 REAL_ADDR_BITS = 16,  ID_BITS = 4, MSG_BITS = 4, OFFSET_BITS = 3) (
					clock, reset, 

					//directory 
					resp_ready, 
					req_ready, 
					in_msg, 
					src_id,
					in_address, 

					out_msg,
					dest_id,
					out_address,

					//memory block 
					readEnable,
					readAddress,
					readData,
					writeEnable,
					writeAddress,
					writeData,

					//main interface 
					cache2mem_msg, 
					cache2mem_address, cache2mem_data,
					mem2cache_msg, 
					mem2cache_address, mem2cache_data
); 
input clock, reset; 

input resp_ready, req_ready; 
input [MSG_BITS-1:0]        in_msg; 
input [ID_BITS-1:0]         src_id;
input [REAL_ADDR_BITS-1:0]  in_address; 
 				     
output [MSG_BITS-1:0]       out_msg; 
output [ID_BITS-1:0]        dest_id;
output [REAL_ADDR_BITS-1:0] out_address; 

output readEnable;
output [REAL_ADDR_BITS-1:0] readAddress;
input [DATA_WIDTH-1:0]      readData;
output writeEnable;
output [REAL_ADDR_BITS-1:0] writeAddress;
output [DATA_WIDTH-1:0]     writeData;

input [MSG_BITS-1:0]        cache2mem_msg; 
input [ADDRESS_BITS-1:0]    cache2mem_address; 
input [DATA_WIDTH-1:0]      cache2mem_data;

output [MSG_BITS-1:0]       mem2cache_msg; 
output [ADDRESS_BITS-1:0]   mem2cache_address; 
output [DATA_WIDTH-1:0]     mem2cache_data;

reg [MSG_BITS-1:0]          mem2cache_msg;
reg [ADDRESS_BITS-1:0]      mem2cache_address; 
reg [DATA_WIDTH-1:0]        mem2cache_data;

reg [MSG_BITS-1:0]          out_msg; 
reg [ID_BITS-1:0]           dest_id;
reg [REAL_ADDR_BITS-1:0]    out_address; 

reg readEnable;
reg [REAL_ADDR_BITS-1:0]    readAddress;
reg writeEnable;
reg [REAL_ADDR_BITS-1:0]    writeAddress;
reg [DATA_WIDTH-1:0]        writeData;
     
localparam OFFSET = 1<<OFFSET_BITS; 
localparam NO_MSG = 0;
localparam M_REQ = 1, INV_REQ = 2, M_WB_REQ = 3, SHARE_REQ = 4, WB_INV_REQ = 5; //memory requests
localparam M_READY = 6, MEM_SENT = 7,  W_RESP = 8, EX_RESP = 9, SH_RESP = 10;   //memory responses 
localparam R_REQ = 1, W_REQ = 2, C_WB_REQ = 3;                                  //cache requests
localparam C_READY = 4, C_SENT = 5, C_RECV = 6, INV_RESP = 7, WB_RESP = 8, WB_INV_RESP = 9;
//cache responses 

localparam VC_DEPTH_BITS = 3; 
localparam INFO_WIDTH = MSG_BITS + ADDRESS_BITS + DATA_WIDTH; 

wire req_write            = ((cache2mem_msg == R_REQ) | (cache2mem_msg == W_REQ));
wire [INFO_WIDTH-1: 0] req_in_data = {cache2mem_msg, cache2mem_address, cache2mem_data};

wire resp_write   = ((cache2mem_msg == SH_RESP) | (cache2mem_msg == INV_RESP) | 
					(cache2mem_msg == WB_RESP) | (cache2mem_msg == WB_INV_RESP)|
					(cache2mem_msg == C_WB_REQ));
wire [INFO_WIDTH-1: 0] resp_in_data = {cache2mem_msg, cache2mem_address, cache2mem_data};

reg  req_read, resp_read;
wire req_valid;
wire [INFO_WIDTH-1: 0] req_data;
wire req_full, req_empty;
  
wire resp_valid;
wire [INFO_WIDTH-1: 0] resp_data;
wire resp_full, resp_empty;
reg [MSG_BITS-1:0]       cache2mem_rq_msg; 
reg [ADDRESS_BITS-1:0]   cache2mem_rq_address; 
reg [DATA_WIDTH-1:0]     cache2mem_rq_data; 

reg [MSG_BITS-1:0]       cache2mem_rp_msg; 
reg [ADDRESS_BITS-1:0]   cache2mem_rp_address; 
reg [DATA_WIDTH-1:0]     cache2mem_rp_data; 

reg [MSG_BITS-1:0]       buf_c2m_rp_msg; 
reg [ADDRESS_BITS-1:0]   buf_c2m_rp_addr; 
reg [DATA_WIDTH-1:0]     buf_c2m_rp_data; 

localparam IDLE = 0, READING = 1, READ_DELAY = 2, READ_OUT = 3, WRITING = 4; 
reg ready, response; 
reg [2:0] state;
reg [OFFSET_BITS :0] counter;

wire no_inflight_msg = ((in_msg == NO_MSG) & (cache2mem_rq_msg == NO_MSG) & 
					(cache2mem_rp_msg == NO_MSG) & (buf_c2m_rp_msg == NO_MSG) & 
					(out_msg == NO_MSG)); 
wire wb_delay = ((cache2mem_rp_msg == WB_RESP) | (cache2mem_rp_msg == WB_INV_RESP));
  
fifo  #(INFO_WIDTH, VC_DEPTH_BITS, 0) req_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(req_in_data),
	.wrtEn(req_write),
	.rdEn(req_read),
	.peek(1'b0),

	.read_data(req_data), 
	.valid(req_valid),			
	.full(req_full), 
	.empty(req_empty) 
);

fifo  #(INFO_WIDTH, VC_DEPTH_BITS, 0) resp_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(resp_in_data),
	.wrtEn(resp_write),
	.rdEn(resp_read),
	.peek(1'b0),

	.read_data(resp_data), 
	.valid(resp_valid),			
	.full(resp_full), 
	.empty(resp_empty) 
);
					
always @(posedge clock) begin 
	if(reset==1)begin 
		cache2mem_rq_msg     <= 0;
		cache2mem_rq_address <= 0;
		cache2mem_rq_data    <= 0; 

		cache2mem_rp_msg     <= 0;
		cache2mem_rp_address <= 0;
		cache2mem_rp_data    <= 0;
		
		req_read             <= 0; 
		resp_read            <= 0;
		response             <= 0; 

	end 
    else begin
		resp_read <= (ready & resp_ready & no_inflight_msg & (~resp_empty) & (~resp_read) & (~response))? 1: 0; 
		response  <= resp_read; 
		req_read  <= (ready & req_ready & no_inflight_msg & resp_empty & (~req_empty) & (~req_read) & (~resp_read) & (~response))? 1: 0; 
		
		cache2mem_rq_msg     <= req_valid? req_data[(INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
		cache2mem_rq_address <= req_valid? req_data[((INFO_WIDTH - MSG_BITS)-1) -: ADDRESS_BITS] : 0; 
		cache2mem_rq_data    <= req_valid? req_data[((INFO_WIDTH - MSG_BITS - ADDRESS_BITS)-1) : 0] : 0; 

		cache2mem_rp_msg     <= resp_valid? resp_data[(INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
		cache2mem_rp_address <= resp_valid? resp_data[((INFO_WIDTH - MSG_BITS)-1) -: ADDRESS_BITS] : 0; 
		cache2mem_rp_data    <= resp_valid? resp_data[((INFO_WIDTH - MSG_BITS - ADDRESS_BITS)-1) : 0] : 0; 
	end 
end 

always @(posedge clock) begin 
	if(reset==1)begin 
		mem2cache_msg     <= 0; 
		mem2cache_address <= 0; 
        mem2cache_data    <= 0; 
		
		readEnable        <= 0;
		readAddress       <= 0;
		writeEnable       <= 0;
		writeData         <= 0;
		
		buf_c2m_rp_msg    <= 0;
		buf_c2m_rp_addr   <= 0;
		buf_c2m_rp_data   <= 0;
		
		ready             <= 1; 
		state             <= IDLE;
	end 
    else begin
		if(ready) begin 
			if (in_msg == NO_MSG) begin 
				if ((cache2mem_rp_msg == WB_RESP) | (cache2mem_rp_msg == WB_INV_RESP)) begin 
					mem2cache_msg     <= M_READY; 
					mem2cache_address <= cache2mem_rp_address;
					mem2cache_data    <= cache2mem_rp_data;		
					 
					buf_c2m_rp_msg    <= cache2mem_rp_msg;
					buf_c2m_rp_addr   <= cache2mem_rp_address;
					buf_c2m_rp_data   <= cache2mem_rp_data;	
					
					ready             <= 0;
					counter           <= 0; 					
					state             <= WRITING;
				end 
				else begin 
					mem2cache_msg     <= NO_MSG; 
					mem2cache_address <= 0; 
					mem2cache_data    <= 0; 
				end 
			end
			else begin 
				if((in_msg == INV_REQ) | (in_msg == WB_INV_REQ) | (in_msg == SHARE_REQ) | (in_msg == M_WB_REQ)) begin 
					mem2cache_msg     <= in_msg; 
					mem2cache_address <= in_address; 
					mem2cache_data    <= src_id; 
				end 
				else begin 
					if((in_msg == W_RESP) | (in_msg == EX_RESP) | (in_msg == SH_RESP)) begin 
						mem2cache_msg     <= in_msg; 
						mem2cache_address <= in_address; 
						mem2cache_data    <= src_id; 
						
						ready             <= 0;
						counter           <= 0;
						readEnable        <= 0;
						readAddress       <= in_address;
						state             <= READING;
					end 
				end 
			end 
		end 
		else begin 
			case (state)
				IDLE: begin
					mem2cache_address <= 0; 
					mem2cache_data    <= 0;
					mem2cache_msg     <= NO_MSG;
					
					ready             <= 1;
					state   		  <= IDLE;
				end
				
				READING: begin
					if((counter == 0) & (cache2mem_msg  == C_READY) & (cache2mem_address == readAddress))begin
						mem2cache_address <= 0; 
						mem2cache_data    <= 0;
						mem2cache_msg     <= NO_MSG;
						
						readEnable        <= 1;
						counter           <= counter + 1;
						state             <= READ_DELAY; 
					end 
					else begin
						if((counter > 0) & (counter < OFFSET) & (cache2mem_msg  == C_RECV)) begin
							mem2cache_address <= 0; 
							mem2cache_data    <= 0;
							mem2cache_msg     <= NO_MSG;
							
							readEnable        <= 1;
							readAddress       <= readAddress + 1;
							counter           <= counter + 1;
							state             <= READ_DELAY;   
						end     
						else begin 
							mem2cache_address <= 0; 
							mem2cache_data    <= 0;
							mem2cache_msg     <= NO_MSG;
							
							readEnable    	  <= 0;
							state             <= READING; 
						end
					end 
				end
							
				READ_DELAY: begin
					readEnable        <= 0;
					state             <= READ_OUT; 
				end 
				
				READ_OUT: begin
					mem2cache_address <= readAddress; 
					mem2cache_data    <= readData;
					mem2cache_msg     <= MEM_SENT;
					
					readEnable        <= 0;
					state             <= (counter == OFFSET)? IDLE : READING; 
				end
				   
				WRITING: begin
					if(counter < OFFSET) begin
						if(cache2mem_msg == C_SENT)begin
							writeAddress   <= cache2mem_address + counter;
							writeData      <= cache2mem_data; 
							writeEnable    <= 1;
							counter        <= counter + 1; 

							mem2cache_msg     <= M_READY; 
							mem2cache_data    <= buf_c2m_rp_data; 							
						end
						else begin
							mem2cache_msg  <= 0; 
							writeEnable    <= 0;
						end        
					end
					else begin
						buf_c2m_rp_msg     <= 0;
						buf_c2m_rp_addr    <= 0;
						buf_c2m_rp_data    <= 0;
						writeEnable        <= 0;
						state              <= (cache2mem_msg == C_SENT)? WRITING : IDLE;  
					end
				end
			endcase 
		end 
	end 
end 
 								
always @ (posedge clock) begin 
	if (reset == 1) begin 
		out_msg     <=  NO_MSG; 
		dest_id     <=  0;
		out_address <=  0; 
	end 
	else  begin
		out_msg     <=  ((state == WRITING) & (counter == OFFSET))? buf_c2m_rp_msg: 
						(~(resp_ready | req_ready | ready))? NO_MSG : 
						(response & (~wb_delay))? cache2mem_rp_msg : cache2mem_rq_msg; 
		dest_id     <=  ((state == WRITING) & (counter == OFFSET))? buf_c2m_rp_data:
						(~(resp_ready | req_ready | ready))? 0 :      
						(response & (~wb_delay))? cache2mem_rp_data    : cache2mem_rq_data;
		out_address <=  ((state == WRITING) & (counter == OFFSET))? buf_c2m_rp_addr:
						(~(resp_ready | req_ready | ready))? 0 :      
						(response & (~wb_delay))? cache2mem_rp_address : cache2mem_rq_address; 
	end
end

/*
always @ (posedge clock) begin 
	$display ("------------------------------ MAIN MEMORY %d-------------------------------------------", CORE); 
	$display ("IN_MSG [%d]\t\t| Src [%d]\t\t| Address [%h]", in_msg, src_id, in_address);
	$display ("OUT_MSG [%d]\t\t| Dest [%d]\t\t| Address [%h]", out_msg, dest_id, out_address);
	$display ("Read [%b]\t\t| Address [%h]\t| Out_data [%h]", readEnable, readAddress, readData);
	$display ("Write [%b]\t\t| Address [%h]\t| In_data [%h]", writeEnable, writeAddress, writeData);
	$display ("");
	$display ("Ready [%b]\t\t| State [%d]\t\t| Counter [%d]", ready, state, counter); 
	$display ("Rp-ready [%b]\t\t| Rq-ready [%b]\t\t| Response [%b]", resp_ready, req_ready, response); 
	$display ("Rp_write [%b]\t\t| Rp_read [%b]\t\t| Rq_write [%b]\t\t| Rq_read [%b]", resp_write, 
				resp_read, req_write, req_read); 
	$display ("Req_valid [%b]\t\t| Req_data [%h]", req_valid, req_data); 
	$display ("Resp_valid [%b]\t| Resp_data [%h]", resp_valid, resp_data); 
	$display ("CAC2MEM MSG-RP [%d]\t| CAC2MEM Add-RP[%h]\t| CACHE2MEM DATA-RP[%h]", 
				 cache2mem_rp_msg, cache2mem_rp_address,cache2mem_rp_data);
	$display ("CAC2MEM MSG-RQ [%d]\t| CAC2MEM Add-RQ[%h]\t| CACHE2MEM DATA-RQ[%h]", 
				 cache2mem_rq_msg, cache2mem_rq_address,cache2mem_rq_data);
	$display ("");		 
	$display ("MEM2CACHE MSG [%d]\t| MEM2CACHE Add [%h]\t| MEM2CACHE DATA [%h]", 
				 mem2cache_msg, mem2cache_address, mem2cache_data);
	$display ("CACHE2MEM MSG [%d]\t| CACHE2MEM Add [%h]\t| CACHE2MEM DATA [%h]", 
				 cache2mem_msg, cache2mem_address,cache2mem_data);
	$display ("------------------------------------------------------------------------------------");
end 
//*/

endmodule 