/** @module : Mem_System_Wrapper
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
 
// Mem_System_Wrapperfor local memory and network 
module directory_mem_system_wrapper #(parameter CORE = 0, ID_BITS = 4, DATA_WIDTH = 32, INDEX_BITS = 6, 
                            OFFSET_BITS = 3, ADDRESS_BITS = 20, SHARERS = 4, INIT_FILE = "memory.mem",
                            IN_PORTS = 1, OUT_PORTS = 1, VC_BITS = 1, VC_DEPTH_BITS = 4,
                            MSG_BITS = 4,EXTRA  = 2, TYPE_BITS = 2)( // report is used to show  
						                                             // performance statistics
             		 clock, reset,
				     i_read, i_write, i_address, i_in_data, 
				     i_out_addr, i_out_data, i_valid, i_ready,
				     
				     d_read, d_write, d_address, d_in_data, 
				     d_out_addr, d_out_data, d_valid, d_ready, 
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full,
					 // performance reporting
					 report 
);

localparam VC_PER_PORTS  = (1 << VC_BITS); 
localparam FLOW_BITS = (2*ID_BITS) + EXTRA;
localparam PAY_WIDTH     = MSG_BITS + DATA_WIDTH;
localparam FLIT_WIDTH    = FLOW_BITS + TYPE_BITS + VC_BITS + PAY_WIDTH;
localparam REAL_ADDR_BITS = ADDRESS_BITS - ID_BITS; 
localparam CACHE_ID_BITS = ID_BITS + 1;

input clock, reset;
input  i_read, i_write;
input  [ADDRESS_BITS-1:0] i_address;
input  [DATA_WIDTH-1:0] i_in_data;

output i_valid, i_ready;
output [DATA_WIDTH-1:0] i_out_data;
output [ADDRESS_BITS-1:0] i_out_addr;

input  d_read, d_write;
input  [ADDRESS_BITS-1:0] d_address;
input  [DATA_WIDTH-1:0] d_in_data;

output d_valid, d_ready;
output [DATA_WIDTH-1:0] d_out_data;
output [ADDRESS_BITS-1:0] d_out_addr;

output [(FLIT_WIDTH * IN_PORTS) -1:0] from_core_flit;
output [IN_PORTS-1: 0] v_from_core;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_empty;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_full; 
  
input  [(FLIT_WIDTH * OUT_PORTS) -1:0] to_core_flit;
input  [OUT_PORTS-1: 0] v_to_core;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_empty;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_full;
input  report;
 
wire [1:0] read  = {d_read, i_read}; 
wire [1:0] write = {d_write, i_write};
wire [(2*ADDRESS_BITS)-1:0] address = {d_address, i_address};
wire [(2*DATA_WIDTH)-1:0] in_data   = {d_in_data, i_in_data};

wire [1:0] valid, ready;
wire[(2*DATA_WIDTH)-1:0] out_data;
wire[(2*ADDRESS_BITS)-1:0] out_addr;

wire [ID_BITS - 1: 0]     core_ID = CORE; 

wire [(2 * MSG_BITS)-1:0]     mem2cache_msg; 
wire [(2 * ADDRESS_BITS)-1:0] mem2cache_address;
wire [(2 * DATA_WIDTH)-1:0]   mem2cache_data;

wire  [(2 * MSG_BITS)-1:0]    cache2mem_msg; 
wire [(2 * ADDRESS_BITS)-1:0] cache2mem_address; 
wire [(2 * DATA_WIDTH)-1:0]   cache2mem_data;

wire [MSG_BITS - 1:0]     cache2local_msg; 
wire [ADDRESS_BITS - 1:0] cache2local_address; 
wire [DATA_WIDTH - 1:0]   cache2local_data;

wire [MSG_BITS - 1:0]     local2cache_msg; 
wire [ADDRESS_BITS - 1:0] local2cache_address; 
wire [DATA_WIDTH - 1:0]   local2cache_data;

wire [MSG_BITS - 1:0]     caches2mem_msg; 
wire [ADDRESS_BITS - 1:0] caches2mem_address; 
wire [DATA_WIDTH - 1:0]   caches2mem_data;

wire [MSG_BITS - 1:0]     mem2caches_msg; 
wire [ADDRESS_BITS - 1:0] mem2caches_address;
wire [DATA_WIDTH - 1:0]   mem2caches_data;

wire [MSG_BITS-1:0]       network2mem_msg; 
wire [ADDRESS_BITS-1:0]   network2mem_address; 
wire [DATA_WIDTH-1:0]     network2mem_data;

wire [MSG_BITS - 1:0]     network2cache_msg; 
wire [ADDRESS_BITS - 1:0] network2cache_address; 
wire [DATA_WIDTH - 1:0]   network2cache_data;

wire [MSG_BITS-1:0]       mem2network_msg;
wire [ADDRESS_BITS-1:0]   mem2network_address;
wire [DATA_WIDTH-1:0]     mem2network_data;

wire [MSG_BITS - 1:0]     cache2network_msg; 
wire [ADDRESS_BITS - 1:0] cache2network_address; 
wire [DATA_WIDTH - 1:0]   cache2network_data;
								
								
unified_caches #(CORE, DATA_WIDTH, INDEX_BITS, OFFSET_BITS, ADDRESS_BITS, MSG_BITS) 
             unified (
				     clock, reset,
				       
				     read, write, address, in_data, 
				     out_addr, out_data, valid, ready,
				     
				     mem2cache_msg, 
				     mem2cache_address, mem2cache_data, 
				     
				     cache2mem_msg, 
					 cache2mem_address, cache2mem_data, 
					 report
); 
cache_splitter #(CORE, DATA_WIDTH, INDEX_BITS, OFFSET_BITS, ADDRESS_BITS, CACHE_ID_BITS, ID_BITS, MSG_BITS )
			C_SPLIT(
                     clock, reset,
					 cache2mem_msg, cache2mem_address, cache2mem_data,
				     mem2cache_msg, mem2cache_address, mem2cache_data, 
					 local2cache_msg, local2cache_address, local2cache_data,
					 cache2local_msg, cache2local_address, cache2local_data,
					 network2cache_msg, network2cache_address, network2cache_data,
					 cache2network_msg, cache2network_address, cache2network_data
 ); 

main_memory #(CORE, DATA_WIDTH, ADDRESS_BITS, CACHE_ID_BITS, SHARERS, REAL_ADDR_BITS, MSG_BITS, OFFSET_BITS, INIT_FILE) 
              m_memory (
				     clock, reset,  
				     				     
				     caches2mem_msg, 
				     caches2mem_address, caches2mem_data,
				     
				     mem2caches_msg, 
				     mem2caches_address, mem2caches_data
);
memory_splitter #(CORE, DATA_WIDTH, INDEX_BITS, OFFSET_BITS, ADDRESS_BITS, CACHE_ID_BITS, ID_BITS, MSG_BITS)
			M_SPLIT(
                     clock, reset,
					 caches2mem_msg,caches2mem_address, caches2mem_data,		 
				     mem2caches_msg, mem2caches_address, mem2caches_data, 				 
					 local2cache_msg, local2cache_address, local2cache_data,
					 cache2local_msg, cache2local_address, cache2local_data,
					 network2mem_msg,  network2mem_address, network2mem_data,
					 mem2network_msg, mem2network_address, mem2network_data
 ); 

packetizer #(CORE, DATA_WIDTH, ADDRESS_BITS, MSG_BITS, OFFSET_BITS, 
             IN_PORTS, OUT_PORTS, VC_BITS, VC_DEPTH_BITS, 
			 CACHE_ID_BITS, ID_BITS, EXTRA, TYPE_BITS) 
			 network_packetizer (
				     clock, reset, 
				     				     
				     network2mem_msg, 
				     network2mem_address, network2mem_data,
				     
				     mem2network_msg, 
				     mem2network_address, mem2network_data,
				     
				     network2cache_msg, 
				     network2cache_address, network2cache_data,
				     
				     cache2network_msg, 
				     cache2network_address, cache2network_data, 
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full 
);

//----------------------------------------------------
// Drive the outputs
//----------------------------------------------------
assign i_valid = valid[0]; 
assign i_ready = ready [0];
assign i_out_data = out_data[DATA_WIDTH-1 :0];
assign i_out_addr = out_addr[ADDRESS_BITS-1 :0];

assign d_valid = valid[1]; 
assign d_ready = ready[1];
assign d_out_data = out_data[(2*DATA_WIDTH)-1 -:DATA_WIDTH];
assign d_out_addr = out_addr[(2*ADDRESS_BITS)-1 -:ADDRESS_BITS];

/*
 always @ (posedge clock) begin 
	$display ("--------------------------------System Wrapper Core %d ----------------------------------", CORE);  
	$display ("cache2mem_msg [%h]\t| cache2mem_address [%h]| cache2mem_data [%h]", cache2mem_msg, cache2mem_address,cache2mem_data);
	$display ("mem2cache_msg [%h]\t| mem2cache_address [%h]| mem2cache_data [%h]", mem2cache_msg, mem2cache_address,mem2cache_data);
	$display ("cache2local_msg [%d]\t| cache2local_address [%h]| cache2local_data [%h]", cache2local_msg, cache2local_address,cache2local_data); 
	$display ("local2cache_msg [%d]\t| local2cache_address [%h]| local2cache_data [%h]", local2cache_msg, local2cache_address,local2cache_data);
	$display ("cache2network_msg [%d]| cache2network_address [%h]| cache2network_data [%h]", cache2network_msg, cache2network_address,cache2network_data);
	$display ("network2cache_msg [%d]| network2cache_address [%h]| network2cache_data [%h]", network2cache_msg, network2cache_address,network2cache_data);

	$display ("caches2mem_msg [%d]\t| caches2mem_address [%h]| caches2mem_data [%h]", caches2mem_msg, caches2mem_address,caches2mem_data);
	$display ("mem2caches_msg [%d]\t| mem2caches_address [%h]| mem2caches_data [%h]", mem2caches_msg, mem2caches_address,mem2caches_data);
	$display ("network2mem_msg [%d]\t| network2mem_address [%h]| network2mem_data [%h]", network2mem_msg, network2mem_address,network2mem_data);
	$display ("mem2network_msg [%d]\t| mem2network_address [%h]| mem2network_data [%h]", mem2network_msg, mem2network_address,mem2network_data);
	$display ("-----------------------------------------------------------------------------------");
 end
//*/

endmodule
