
/** @module : Dir_packetizer
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

module packetizer #(parameter CORE = 0, DATA_WIDTH = 32, ADDRESS_BITS = 20, MSG_BITS = 4, 
					OFFSET_BITS= 3, IN_PORTS = 1, OUT_PORTS = 1, VC_BITS = 1, 
                    VC_DEPTH_BITS = 4, CACHE_ID_BITS = 5, 
					ID_BITS = 4, EXTRA  = 2, TYPE_BITS = 2) ( 
					
				     clock, reset, 
				     				     
				     network2mem_msg, 
				     network2mem_address, network2mem_data,
				     
				     mem2network_msg, 
				     mem2network_address, mem2network_data,
				     
				     network2cache_msg, 
				     network2cache_address, network2cache_data,
				     
				     cache2network_msg, 
				     cache2network_address, cache2network_data,
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full  
				     
);
localparam VC_PER_PORTS  = (1 << VC_BITS); 
localparam FLOW_BITS     = (2*ID_BITS) + EXTRA;  
localparam PAY_WIDTH     = MSG_BITS + DATA_WIDTH;
localparam FLIT_WIDTH    = FLOW_BITS + TYPE_BITS + VC_BITS + PAY_WIDTH;
localparam OFFSETS       = 1<<OFFSET_BITS; 
localparam LINE_BITS = (DATA_WIDTH * OFFSETS) + CACHE_ID_BITS; 
input clock, reset; 

input [MSG_BITS-1:0]     mem2network_msg ;
input [ADDRESS_BITS-1:0] mem2network_address ;
input [DATA_WIDTH-1:0]   mem2network_data ;

input [MSG_BITS -1:0]     cache2network_msg; 
input [ADDRESS_BITS -1:0] cache2network_address; 
input [DATA_WIDTH -1:0]   cache2network_data ;

output [MSG_BITS-1:0]     network2mem_msg; 
output [ADDRESS_BITS-1:0] network2mem_address; 
output [DATA_WIDTH-1:0]   network2mem_data;

output [MSG_BITS -1:0]     network2cache_msg; 
output [ADDRESS_BITS -1:0] network2cache_address; 
output [DATA_WIDTH -1:0]   network2cache_data;

output [(FLIT_WIDTH * IN_PORTS) -1:0] from_core_flit;
output [IN_PORTS-1: 0] v_from_core;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_empty;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_full; 
  
input  [(FLIT_WIDTH * OUT_PORTS) -1:0] to_core_flit;
input  [OUT_PORTS-1: 0] v_to_core;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_empty;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_full;

wire                    m_ready; 
wire [MSG_BITS-1:0]     m_in_msg, m_out_msg; 
wire [ADDRESS_BITS-1:0] m_in_address, m_out_address;
wire [LINE_BITS-1:0]    m_in_cache_line, m_out_cache_line;

wire [MSG_BITS-1:0]     c_in_msg, c_out_msg; 
wire [ADDRESS_BITS-1:0] c_in_address, c_out_address;
wire [LINE_BITS-1:0]    c_in_cache_line, c_out_cache_line;

packetizer_mem_emulator #(CORE, DATA_WIDTH, ADDRESS_BITS, CACHE_ID_BITS, 
					 MSG_BITS, OFFSET_BITS) m_emul (
				     clock, reset,  
				     				     
				     cache2network_msg, 
				     cache2network_address, cache2network_data,
				     network2cache_msg, 
				     network2cache_address, network2cache_data, 
				     
					 m_in_msg,
				     m_in_address, 
         		  	 m_in_cache_line,
            		
					 m_ready, 
					 m_out_msg,
         		 	 m_out_address,
         		  	 m_out_cache_line
); 

packetizer_cache_emulator #(CORE, DATA_WIDTH, ADDRESS_BITS, CACHE_ID_BITS, 
					MSG_BITS, OFFSET_BITS) c_emul (
				     clock, reset,  
				   
				     mem2network_msg, 
				     mem2network_address, mem2network_data, 
				     
				     network2mem_msg, 
				     network2mem_address, network2mem_data, 
				      
					 c_in_msg,
				     c_in_address, 
         		  	 c_in_cache_line,
            		
					 c_out_msg,
         		 	 c_out_address,
         		  	 c_out_cache_line
);
					
network_interface #(CORE, DATA_WIDTH, ADDRESS_BITS, MSG_BITS, OFFSET_BITS, 
                         IN_PORTS, OUT_PORTS, VC_BITS, VC_DEPTH_BITS, 
						 CACHE_ID_BITS, ID_BITS, EXTRA, TYPE_BITS) net_inter (
				     .clock(clock), 
					 .reset(reset),   		     				     
				     
				     .from_core_flit(from_core_flit), 
					 .v_from_core(v_from_core), 
				     .from_core_empty(from_core_empty), 
					 .from_core_full(from_core_full),
  
  					 .to_core_flit(to_core_flit), 
					 .v_to_core(v_to_core),
  					 .to_core_empty(to_core_empty), 
					 .to_core_full(to_core_full),  
				     
					 .me_ready(m_ready),
					 .me_in_msg(m_out_msg), 
				     .me_in_address(m_out_address),  
         		  	 .me_in_cache_line(m_out_cache_line), 
            		 
					 .me_out_msg(m_in_msg), 
         		 	 .me_out_address(m_in_address), 
         		  	 .me_out_cache_line(m_in_cache_line), 
					 
					 .ce_in_msg(c_out_msg), 
				     .ce_in_address(c_out_address),  
         		  	 .ce_in_cache_line(c_out_cache_line), 
            		
					 .ce_out_msg(c_in_msg), 
         		 	 .ce_out_address(c_in_address), 
         		  	 .ce_out_cache_line(c_in_cache_line)	 
);

/*
 always @ (posedge clock) begin 
	$display ("----------------------------- Packetizer %d-------------------------------", CORE);  
	$display ("network2mem_msg [%d]\t| network2mem_address [%h]  | network2mem_data [%h]", network2mem_msg, network2mem_address,network2mem_data);
	$display ("mem2network_msg [%d]\t| mem2network_address [%h]  | mem2network_data [%h]", mem2network_msg, mem2network_address,mem2network_data);
	$display ("cache2network_msg [%d]| cache2network_address [%h]| cache2network_data [%h]", cache2network_msg, cache2network_address,cache2network_data);
	$display ("network2cache_msg [%d]| network2cache_address [%h]| network2cache_data [%h]", network2cache_msg, network2cache_address,network2cache_data);
	$display (""); 
	$display ("From Core Flit [%h] ", from_core_flit);
	$display ("From Core Valid [%b] ", v_from_core);
	$display ("From Core Empty [%b]\t| From Core Full [%b] ", from_core_empty, from_core_full);
	$display ("To Core Flit [%h] ", to_core_flit);
	$display ("To Core Valid [%b] ", v_to_core);
	$display ("To Core Empty [%b]\t| To Core Full [%b] ", to_core_empty, to_core_full);
	$display ("-----------------------------------------------------------------------------------");
 end 	
//*/

endmodule