/** @module : Packetizer memory emulator
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2013 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
 
module packetizer_mem_emulator #(parameter CORE = 0, DATA_WIDTH = 32, ADDRESS_BITS = 20, ID_BITS = 4, MSG_BITS = 3, OFFSET_BITS= 3) (
				     clock, reset,  
				     				     
				     cache2network_msg, 
				     cache2network_address, cache2network_data,
				     network2cache_msg, 
				     network2cache_address, network2cache_data, 
				     
					 m_in_msg,
				     m_in_address, 
         		  	 m_in_cache_line,
            		
					 me_ready, 
					 m_out_msg,
         		 	 m_out_address,
         		  	 m_out_cache_line
); 

localparam OFFSETS = 1<<OFFSET_BITS; 
localparam LINE_BITS = (DATA_WIDTH * OFFSETS) + ID_BITS; 
localparam NO_MSG = 0;
localparam M_REQ = 1, INV_REQ = 2, M_WB_REQ = 3, SHARE_REQ = 4, WB_INV_REQ = 5; //memory requests
localparam M_READY = 6, MEM_SENT = 7,  W_RESP = 8, EX_RESP = 9, SH_RESP = 10;   //memory responses 
localparam R_REQ = 1, W_REQ = 2, C_WB_REQ = 3;                                  //cache requests
localparam C_READY = 4, C_SENT = 5, C_RECV = 6, INV_RESP = 7, WB_RESP = 8, WB_INV_RESP = 9;
//cache responses 
localparam IDLE = 0, WRITING = 1, READING = 2;
localparam VC_DEPTH_BITS = 3; 
localparam INFO_WIDTH = MSG_BITS + ADDRESS_BITS + DATA_WIDTH; 

input clock, reset; 				     
input [MSG_BITS -1:0]     cache2network_msg; 
input [ADDRESS_BITS -1:0] cache2network_address; 
input [DATA_WIDTH -1:0]   cache2network_data;

output [MSG_BITS -1:0]     network2cache_msg; 
output [ADDRESS_BITS -1:0] network2cache_address; 
output [DATA_WIDTH -1:0]   network2cache_data;

input [MSG_BITS -1:0]    m_in_msg;
input [ADDRESS_BITS-1:0] m_in_address;
input [LINE_BITS-1:0]    m_in_cache_line;
 
output                    me_ready; 
output [MSG_BITS -1:0]    m_out_msg;
output [ADDRESS_BITS-1:0] m_out_address;
output [LINE_BITS-1:0]    m_out_cache_line;

wire [(DATA_WIDTH * OFFSETS)-1:0]  out_cache_line;
reg  [DATA_WIDTH-1:0]   write_word [0:OFFSETS-1];
reg  [(DATA_WIDTH * OFFSETS)-1:0]  in_cache_line;
wire [DATA_WIDTH-1:0]  read_word [0:OFFSETS-1];
reg [OFFSET_BITS :0]    counter; 

reg [MSG_BITS -1:0]     network2cache_msg; 
reg [ADDRESS_BITS -1:0] network2cache_address; 
reg [DATA_WIDTH -1:0]   network2cache_data;
  
reg [MSG_BITS -1:0]    m_out_msg;
reg [ADDRESS_BITS-1:0] m_out_address;
reg [LINE_BITS-1:0]    m_out_cache_line;
wire [(DATA_WIDTH * OFFSETS)-1:0]  padding = 0; 

reg [2:0] state;
wire req_write            = ((cache2network_msg == R_REQ) | (cache2network_msg == W_REQ));
wire [INFO_WIDTH-1: 0] req_in_data = {cache2network_msg, cache2network_address, cache2network_data};

wire resp_write   = ((cache2network_msg == SH_RESP) | (cache2network_msg == INV_RESP) | 
					(cache2network_msg == WB_RESP) | (cache2network_msg == WB_INV_RESP)|
					(cache2network_msg == C_WB_REQ));
wire [INFO_WIDTH-1: 0] resp_in_data = {cache2network_msg, cache2network_address, cache2network_data};

reg  req_read, resp_read;
wire req_valid;
wire [INFO_WIDTH-1: 0] req_data;
wire req_full, req_empty;
  
wire resp_valid;
wire [INFO_WIDTH-1: 0] resp_data;
wire resp_full, resp_empty;
reg [MSG_BITS-1:0]       cache2network_rq_msg; 
reg [ADDRESS_BITS-1:0]   cache2network_rq_address; 
reg [DATA_WIDTH-1:0]     cache2network_rq_data; 

reg [MSG_BITS-1:0]       cache2network_rp_msg; 
reg [ADDRESS_BITS-1:0]   cache2network_rp_address; 
reg [DATA_WIDTH-1:0]     cache2network_rp_data; 

reg [MSG_BITS-1:0]       buf_c2n_rp_msg; 
reg [ADDRESS_BITS-1:0]   buf_c2n_rp_addr; 
reg [DATA_WIDTH-1:0]     buf_c2n_rp_data; 
reg ready, response;  

wire no_inflight_msg = ((m_in_msg == NO_MSG) & (cache2network_rq_msg == NO_MSG) & 
					   (cache2network_rp_msg == NO_MSG) & (buf_c2n_rp_msg == NO_MSG) & 
					   (m_out_msg == NO_MSG)); 
wire wb_delay = ((cache2network_rp_msg == WB_RESP) | (cache2network_rp_msg == WB_INV_RESP));

assign me_ready = ready;

genvar i;
generate   
	for (i = 0; i < OFFSETS; i=i+1) begin : CN_CONNECT
		assign out_cache_line[(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH]= write_word[i];
		assign read_word[i]     = in_cache_line[(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH];
	end 
endgenerate
  
fifo  #(INFO_WIDTH, VC_DEPTH_BITS, 0) req_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(req_in_data),
	.wrtEn(req_write),
	.rdEn(req_read),
	.peek(1'b0),

	.read_data(req_data), 
	.valid(req_valid),			
	.full(req_full), 
	.empty(req_empty) 
);

fifo  #(INFO_WIDTH, VC_DEPTH_BITS, 0) resp_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(resp_in_data),
	.wrtEn(resp_write),
	.rdEn(resp_read),
	.peek(1'b0),

	.read_data(resp_data), 
	.valid(resp_valid),			
	.full(resp_full), 
	.empty(resp_empty) 
);
					
always @(posedge clock) begin 
	if(reset==1)begin 
		cache2network_rq_msg     <= 0;
		cache2network_rq_address <= 0;
		cache2network_rq_data    <= 0; 

		cache2network_rp_msg     <= 0;
		cache2network_rp_address <= 0;
		cache2network_rp_data    <= 0;
		
		req_read                 <= 0; 
		resp_read                <= 0;
		response                 <= 0; 

	end 
    else begin
		resp_read <= (ready & no_inflight_msg & (~resp_empty) & (~resp_read) & (~response))? 1: 0; 
		response  <= resp_read; 
		req_read  <= (ready & no_inflight_msg & resp_empty & (~req_empty) & (~req_read) & 
		             (~resp_read) & (~response))? 1: 0; 
		
		cache2network_rq_msg     <= req_valid? req_data[(INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
		cache2network_rq_address <= req_valid? req_data[((INFO_WIDTH - MSG_BITS)-1) -: ADDRESS_BITS] : 0; 
		cache2network_rq_data    <= req_valid? req_data[((INFO_WIDTH - MSG_BITS - ADDRESS_BITS)-1) : 0] : 0; 

		cache2network_rp_msg     <= resp_valid? resp_data[(INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
		cache2network_rp_address <= resp_valid? resp_data[((INFO_WIDTH - MSG_BITS)-1) -: ADDRESS_BITS] : 0; 
		cache2network_rp_data    <= resp_valid? resp_data[((INFO_WIDTH - MSG_BITS - ADDRESS_BITS)-1) : 0] : 0; 
	end 
end 

always @(posedge clock) begin 
	if(reset==1)begin 
		network2cache_msg     <= 0; 
		network2cache_address <= 0; 
        network2cache_data    <= 0; 
		
		buf_c2n_rp_msg    <= 0;
		buf_c2n_rp_addr   <= 0;
		buf_c2n_rp_data   <= 0;
		
		ready             <= 1; 
		counter           <= 0;
		state             <= IDLE;
	end 
    else begin
		if(ready) begin 
			if (m_in_msg == NO_MSG) begin 
				if ((cache2network_rp_msg == WB_RESP) | (cache2network_rp_msg == WB_INV_RESP)) begin 
					network2cache_msg     <= M_READY; 
					network2cache_address <= cache2network_rp_address;
					network2cache_data    <= cache2network_rp_data;		
					 
					buf_c2n_rp_msg    <= cache2network_rp_msg;
					buf_c2n_rp_addr   <= cache2network_rp_address;
					buf_c2n_rp_data   <= cache2network_rp_data;	
					
					ready             <= 0;
					counter           <= 0; 					
					state             <= WRITING;
				end 
				else begin 
					network2cache_msg     <= NO_MSG; 
					network2cache_address <= 0; 
					network2cache_data    <= 0; 
				end 
			end
			else begin 
				if((m_in_msg == INV_REQ) | (m_in_msg == WB_INV_REQ) | (m_in_msg == SHARE_REQ) | (m_in_msg == M_WB_REQ)) begin 
					network2cache_msg     <= m_in_msg; 
					network2cache_address <= m_in_address; 
					network2cache_data    <= m_in_cache_line[(LINE_BITS -1) -: ID_BITS]; 
				end 
				else begin 
					if((m_in_msg == W_RESP) | (m_in_msg == EX_RESP) | (m_in_msg == SH_RESP)) begin 
						network2cache_msg     <= m_in_msg; 
						network2cache_address <= m_in_address; 
						network2cache_data    <= m_in_cache_line[(LINE_BITS -1) -: ID_BITS]; 
						
						ready             <= 0;
						counter           <= 0;
						in_cache_line     <= m_in_cache_line;
						state             <= READING;
					end 
				end 
			end 
		end 
		else begin 
			case (state)
				IDLE: begin
					network2cache_address <= 0; 
					network2cache_data    <= 0;
					network2cache_msg     <= NO_MSG;
					
					ready                 <= 1;
					state   		      <= IDLE;
				end
				
				READING: begin
					if((counter < OFFSETS) & ((cache2network_msg  == C_RECV) | (cache2network_msg  == C_READY))) begin
						network2cache_msg     <= MEM_SENT; 
						network2cache_address <= cache2network_address + counter; 
						network2cache_data    <= read_word[counter]; 	
						
						counter               <= counter + 1;
						state                 <= READING;   
					end     
					else begin
						network2cache_msg     <= NO_MSG; 
						network2cache_address <= 0; 
						network2cache_data    <= 0; 
						state                 <= (counter == OFFSETS)? IDLE : READING;
					end
				end
					
			
				WRITING: begin
					if(counter < OFFSETS) begin
						if(cache2network_msg == C_SENT)begin
							write_word[counter]   <= cache2network_data; 
							counter               <= counter + 1; 

							network2cache_msg     <= M_READY; 
							network2cache_address <= buf_c2n_rp_addr; 
							network2cache_data    <= buf_c2n_rp_data; 							
						end
						else begin
							network2cache_msg     <= NO_MSG; 
						end  
						state                     <= WRITING; 					
					end
					else begin
						buf_c2n_rp_msg            <= NO_MSG;
						buf_c2n_rp_addr           <= 0;
						buf_c2n_rp_data           <= 0;
						
						state                     <= (cache2network_msg == C_SENT)? WRITING : IDLE;  
					end
				end				
			endcase 
		end 
	end 
end 
				
always @ (posedge clock) begin 
	if (reset == 1) begin 
		m_out_msg          <=  NO_MSG; 
		m_out_address      <=  0;
		m_out_cache_line   <=  0; 
	end 
	else  begin
		m_out_msg          <=  ((state == WRITING) & (counter == OFFSETS))? buf_c2n_rp_msg: (~ready)? NO_MSG : 
						       (response & (~wb_delay))? cache2network_rp_msg : cache2network_rq_msg; 
		m_out_address      <=  ((state == WRITING) & (counter == OFFSETS))? buf_c2n_rp_addr:(~ready)? 0 :  
							   (response & (~wb_delay))? cache2network_rp_address : cache2network_rq_address;		
		m_out_cache_line   <=  ((state == WRITING) & (counter == OFFSETS))? {buf_c2n_rp_data[ID_BITS-1 : 0], out_cache_line}: (~ready)? 0 :      
						       (response & (~wb_delay))? {cache2network_rp_data[ID_BITS-1 : 0], padding} : 
							   {cache2network_rq_data[ID_BITS-1 : 0], padding}; 
	end
end


/*
always @ (posedge clock) begin 
	$display ("------------------------------ Memory Emulator -------------------------------------------"); 
	$display ("IN_MSG [%d]\t\t| Dest [%d]\t\t| Address [%h]", m_in_msg, m_in_cache_line[(LINE_BITS -1) -: ID_BITS], m_in_address);
	$display ("OUT_MSG [%d]\t\t| Src [%d]\t\t| Address [%h]", m_out_msg, m_out_cache_line[(LINE_BITS -1) -: ID_BITS], m_out_address);
	$display ("");
	$display ("Ready [%b]\t\t| Response [%b]\t\t| State [%d]\t\t| Counter [%d]", ready, response, state, counter); 
	$display ("Rp_write [%b]\t\t| Rp_read [%b]\t\t| Rq_write [%b]\t\t| Rq_read [%b]", resp_write, 
				resp_read, req_write, req_read); 
	$display ("Req_valid [%b]\t\t| Req_data [%h]", req_valid, req_data); 
	$display ("Resp_valid [%b]\t\t| Resp_data [%h]", resp_valid, resp_data); 
	$display ("CAC2NET MSG-RP [%d]\t| CAC2NET Add-RP[%h]\t| CACHE2NET DATA-RP[%h]", 
				 cache2network_rp_msg, cache2network_rp_address,cache2network_rp_data);
	$display ("CAC2NET MSG-RQ [%d]\t| CAC2NET Add-RQ[%h]\t| CACHE2NET DATA-RQ[%h]", 
				 cache2network_rq_msg, cache2network_rq_address,cache2network_rq_data);
	$display ("");		 
	$display ("NET2CACHE MSG [%d]\t| NET2CACHE Add [%h]\t| NET2CACHE DATA [%h]", 
				 network2cache_msg, network2cache_address, network2cache_data);
	$display ("CACHE2NET MSG [%d]\t| CACHE2NET Add [%h]\t| CACHE2NET DATA [%h]", 
				 cache2network_msg, cache2network_address,cache2network_data);
	$display ("------------------------------------------------------------------------------------");
end 
//*/
endmodule 